import os
import json
import subprocess
import uuid
from pathlib import Path
from typing import Dict
import importlib

import modal
import modal.experimental

from app import model_dir, MODEL_VOL, comfyui_out_dir, COMFYUI_OUT_VOL


# if use nightly, git pull comfyui repo
VERSION = os.getenv("VERSION", "0.3.75")

IMAGE_GPU = os.getenv("IMAGE_GPU", "L4")
MAX_INPUTS = int(os.getenv("MAX_INPUTS", "2"))

MODEL_NAME = os.getenv("MODEL_NAME", "flux1_schnell_fp8")
WORKFLOW_CONFIG_DIR = Path(__file__).parent / "workflow_config"
WORKFLOW_CONFIG_PATH = os.getenv(
    "WORKFLOW_CONFIG_PATH", f"{WORKFLOW_CONFIG_DIR}/{MODEL_NAME}_api.json"
)

APP_DIR = Path(__file__).parent / "app"

image = (  # build up a Modal Image to run ComfyUI, step by step
    modal.Image.debian_slim(  # start from basic Linux with Python
        python_version="3.11"
    )
    .apt_install("git")  # install git to clone ComfyUI
    .uv_pip_install("fastapi[standard]==0.115.4")  # install web dependencies
    # copy the app path to the container.
    .add_local_dir(APP_DIR, f"/root/app", copy=True)
    # copy the ComfyUI workflow JSON to the container.
    .add_local_file(
        WORKFLOW_CONFIG_PATH,
        f"/root/{MODEL_NAME}_api.json",
        copy=True,
    )
    .env(
        {
            "MODEL_NAME": MODEL_NAME,
        }
    )
)

image = (
    # https://github.com/Comfy-Org/comfy-cli
    image.uv_pip_install("comfy-cli==1.5.3")  # install comfy-cli
    .run_commands(  # use comfy-cli to install ComfyUI and its dependencies
        # https://github.com/comfyanonymous/ComfyUI
        f"comfy --skip-prompt install --fast-deps --nvidia --version {VERSION}"
    )
    .run_commands(
        # https://github.com/WASasquatch/was-node-suite-comfyui
        "comfy node install --fast-deps was-ns@3.0.1"
    )
    .run_commands(
        "which comfy",
        "comfy --help",
    )
)


# https://docs.comfy.org/installation/manual_install#example-structure
def link_comfyui_dir():
    try:
        MODEL_NAME = os.getenv("MODEL_NAME")
        module = importlib.import_module(f"app.{MODEL_NAME}")
        func = getattr(module, "link_comfyui_dir")
        func()
    except ImportError as e:
        print(e)
    except Exception as e:
        print(e)


image = image.run_function(
    link_comfyui_dir,
    volumes={model_dir: MODEL_VOL},
)


app = modal.App(name="server", image=image)


@app.cls(
    scaledown_window=300,  # 5 minute container keep alive after it processes an input
    gpu=IMAGE_GPU,
    volumes={
        model_dir: MODEL_VOL,
        comfyui_out_dir: COMFYUI_OUT_VOL,
    },
    timeout=1200,  # default 300s
)
@modal.concurrent(max_inputs=MAX_INPUTS)  # run max inputs per container
class ComfyUI:
    port: int = 8000

    @modal.enter()
    def launch_comfy_background(self):
        # launch the ComfyUI server exactly once when the container starts
        cmd = f"comfy launch --background -- --port {self.port}"
        subprocess.run(cmd, shell=True, check=True)

    @modal.method()
    def infer(self, file_name: str = ""):
        # sometimes the ComfyUI server stops responding (we think because of memory leaks), so this makes sure it's still up
        self.poll_server_health()

        # runs the comfy run --workflow command as a subprocess
        workflow_path = f"{file_name}.json"
        # workflow_path = Path(__file__).parent / f"{os.getenv("MODEL_NAME")}_api.json"
        cmd = f"comfy run --workflow {workflow_path} --wait --timeout 1200 --verbose"
        print(cmd)
        try:
            result = subprocess.run(
                cmd,
                shell=True,
                check=True,
                capture_output=True,
                text=True,  # Decode stdout/stderr as text
            )
            print("ComfyUI stdout:", result.stdout)
            if result.stderr:
                print("ComfyUI stderr:", result.stderr)
        except subprocess.CalledProcessError as e:
            print(f"ComfyUI command failed with exit code {e.returncode}")
            print(f"ComfyUI stdout: {e.stdout}")
            print(f"ComfyUI stderr: {e.stderr}")
            raise  # Re-raise the exception after logging

    @modal.asgi_app()
    def api(self):
        from fastapi import FastAPI, Response
        from fastapi.responses import StreamingResponse

        app = FastAPI()

        @app.post(path="/video")
        def gen_video(item: Dict):
            model_name = item.get("model")
            if model_name is None:
                return Response("Missing model name", status_code=400)

            MODEL_NAME = os.getenv("MODEL_NAME")
            if model_name != MODEL_NAME:
                return Response(f"Model name does not match, now use {MODEL_NAME}", status_code=400)

            file_path = Path(__file__).parent / f"{MODEL_NAME}_api.json"
            images = item.get("images")
            if images is not None and isinstance(images, list) and len(images) > 0:
                file_path = Path(__file__).parent / f"{MODEL_NAME}_img2img_api.json"
            # change workflow conf
            filename = self.change_workflow_conf(file_path, item)
            if filename is None:
                return Response("Failed to change workflow conf", status_code=500)

            # run inference on the currently running container
            # NOTE: need use modal queue to do this
            self.infer.local(filename)

            # returns the video as bytes stream
            # https://fastapi.tiangolo.com/advanced/custom-response/#using-streamingresponse-with-file-like-objects
            def iterfile():  # (1)
                for f in Path(comfyui_out_dir).iterdir():
                    if f.name.startswith(filename):
                        # if f.name.endswith(".mp4"):
                        with open(f, mode="rb") as file_like:  # (2)
                            yield from file_like  # (3)

            return StreamingResponse(iterfile(), media_type="video/mp4")

        @app.post(path="/image")
        def gen_image(item: Dict):
            model_name = item.get("model")
            if model_name is None:
                return Response("Missing model name", status_code=400)

            MODEL_NAME = os.getenv("MODEL_NAME")
            if model_name != MODEL_NAME:
                return Response(f"Model name does not match, now use {MODEL_NAME}", status_code=400)

            file_path = Path(__file__).parent / f"{MODEL_NAME}_api.json"
            images = item.get("images")
            if images is not None and isinstance(images, list) and len(images) > 0:
                file_path = Path(__file__).parent / f"{MODEL_NAME}_img2img_api.json"
            # change workflow conf
            filename = self.change_workflow_conf(file_path, item)
            if filename is None:
                return Response("Failed to change workflow conf", status_code=500)

            # run inference on the currently running container
            self.infer.local(filename)

            # returns the image as bytes
            img_bytes = b""
            for f in Path(comfyui_out_dir).iterdir():
                if f.name.startswith(filename):
                    img_bytes = f.read_bytes()

            return Response(img_bytes, media_type="image/jpeg")

        return app

    def change_workflow_conf(self, file_path: Path, item: Dict) -> str:
        try:
            MODEL_NAME = os.getenv("MODEL_NAME")
            module = importlib.import_module(f"app.{MODEL_NAME}")
            func = getattr(module, "change_workflow_conf")
            filename_prefix = func(file_path, **item)
            return filename_prefix
        except ImportError as e:
            print(f"Import error in change_workflow_conf: {e}")
            return None
        except Exception as e:
            print(f"An unexpected error occurred in change_workflow_conf: {e}")
            return None

    def poll_server_health(self) -> None:
        import socket
        import urllib

        try:
            # check if the server is up (response should be immediate)
            req = urllib.request.Request(f"http://127.0.0.1:{self.port}/system_stats")
            urllib.request.urlopen(req, timeout=5)
            print("ComfyUI server is healthy")
        except (socket.timeout, urllib.error.URLError) as e:
            # if no response in 5 seconds, stop the container
            print(f"Server health check failed: {str(e)}")
            modal.experimental.stop_fetching_inputs()

            # all queued inputs will be marked "Failed", so you need to catch these errors in your client and then retry
            raise Exception("ComfyUI server is not healthy, stopping container")


"""
# flux1.0 schnell fp8
modal serve src/comfyui/server.py 
MODEL_NAME=flux1_schnell_fp8 IMAGE_GPU=L40S modal serve src/comfyui/server.py 

# z-image turbo
MODEL_NAME=image_z_image_turbo IMAGE_GPU=L4 modal serve src/comfyui/server.py 
MODEL_NAME=image_z_image_turbo IMAGE_GPU=L40S modal serve src/comfyui/server.py 

# flux2 dev
MODEL_NAME=image_flux2 IMAGE_GPU=L40S modal serve src/comfyui/server.py 

# hunyuan video 1.5 720p t2v
MODEL_NAME=video_hunyuan_video_1_5_720p_t2v IMAGE_GPU=L40S modal serve src/comfyui/server.py
"""



from dataclasses import dataclass, field
from typing import List

from achatbot.types.llm.transformers import TransformersLMArgs


@dataclass
class VitaAudioTokenizerArgs:
    audio_tokenizer_type: str = "sensevoice_glm4voice"  # glm4voice | sensevoice_glm4voice
    audio_tokenizer_model_path: str = None
    sense_voice_model_path: str = None
    flow_path: str = None
    audio_tokenizer_rank: int = 0
    chunk_size_list: list = field(default_factory=lambda: [25, 50, 100, 150, 200])


@dataclass
class VitaAudioTransformersVoiceLMArgs(TransformersLMArgs, VitaAudioTokenizerArgs):
    """ """



import os

import dotenv
from mem0 import Memory, MemoryClient
from langchain_openai import ChatOpenAI
from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import InMemoryVectorStore, Chroma

dotenv.load_dotenv()

# zhipu GLM
llm = ChatOpenAI(
    base_url="https://open.bigmodel.cn/api/paas/v4",
    model="glm-4.5-flash",
    max_tokens=32768,
    temperature=0.2,
    api_key=os.environ["ZHIPU_API_KEY"],
)


# Set embeddings
# https://ai.gitee.com/serverless-api#embedding-rerank
# 100 api calls per day, free tier
embd = OpenAIEmbeddings(
    base_url="https://ai.gitee.com/v1",
    model="Qwen3-Embedding-8B",  # 4096
    api_key=os.environ["GITEE_API_KEY"],
    dimensions=1536,
    check_embedding_ctx_length=False,
    chunk_size=1000,
)

# vector_store = Chroma(
#    persist_directory="./rag_chroma_db",
#    embedding_function=embd,
#    collection_name="mem0",  # Required collection name
# )

vector_store = InMemoryVectorStore(embedding=embd)


def langchain_mem0():
    # Pass the initialized model to the config
    # RAG
    config = {
        "llm": {"provider": "langchain", "config": {"model": llm}},
        "embedder": {"provider": "langchain", "config": {"model": embd}},
        # "vector_store": {
        #    "provider": "langchain",
        #    "config": {"client": vector_store},
        # },  # default qdrant
        # "graph_store": None,
        # "ranker": None,
    }

    m = Memory.from_config(config)
    messages = [
        {"role": "user", "content": "I'm planning to watch a movie tonight. Any recommendations?"},
        {"role": "assistant", "content": "How about thriller movies? They can be quite engaging."},
        {
            "role": "user",
            "content": "I'm not a big fan of thriller movies but I love sci-fi movies.",
        },
        {
            "role": "assistant",
            "content": "Got it! I'll avoid thriller recommendations and suggest sci-fi movies in the future.",
        },
    ]
    m.add(
        messages,
        user_id="alice",
        agent_id="achatbot-agent1",
        run_id="bot1",
        metadata={"category": "movies"},
    )
    res = m.search(
        query="Can you suggest some good sci-fi movies?",
        user_id="alice",
        agent_id="achatbot-agent1",
        run_id="bot1",
        limit=3,
    )
    print(res)


def remote_api_mem0():
    m = MemoryClient(
        api_key=os.environ["MEM0_API_KEY"],
        host=None,  # https://api.mem0.ai
    )

    messages = [
        {"role": "user", "content": "I'm planning to watch a movie tonight. Any recommendations?"},
        {"role": "assistant", "content": "How about thriller movies? They can be quite engaging."},
        {
            "role": "user",
            "content": "I'm not a big fan of thriller movies but I love sci-fi movies.",
        },
        {
            "role": "assistant",
            "content": "Got it! I'll avoid thriller recommendations and suggest sci-fi movies in the future.",
        },
    ]
    m.add(
        messages,
        user_id="alice",
        agent_id="achatbot-agent1",
        run_id="bot1",
        metadata={"category": "movies"},
    )

    filters = {
        "OR": [
            {"user_id": "alice"},
            {"agent_id": "achatbot-agent1"},
            {"run_id": "bot1"},
        ]
    }
    res = m.search(
        query="Can you suggest some good sci-fi movies?",
        filters=filters,
        version="v2",
        top_k=3,
        threshold=0.1,
        output_format="v1.1",
    )
    print(res)


"""
MOD=local python -m demo.langchain.mem.mem0
MOD=api python -m demo.langchain.mem.mem0
"""
if __name__ == "__main__":
    mod = os.getenv("MOD", "local")
    if mod == "api":
        remote_api_mem0()
    else:
        langchain_mem0()
 


#!/usr/bin/env python3
import argparse
import logging
from pathlib import Path
import sys
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import torch
from torch.nn.parallel import data_parallel
from typeguard import check_argument_types

from funasr_local.utils.cli_utils import get_commandline_args
from funasr_local.fileio.datadir_writer import DatadirWriter
from funasr_local.tasks.lm import LMTask
from funasr_local.torch_utils.device_funcs import to_device
from funasr_local.torch_utils.forward_adaptor import ForwardAdaptor
from funasr_local.torch_utils.set_all_random_seed import set_all_random_seed
from funasr_local.utils import config_argparse
from funasr_local.utils.types import float_or_none
from funasr_local.utils.types import str2bool
from funasr_local.utils.types import str2triple_str
from funasr_local.utils.types import str_or_none


def calc_perplexity(
    output_dir: str,
    batch_size: int,
    dtype: str,
    ngpu: int,
    seed: int,
    num_workers: int,
    log_level: Union[int, str],
    data_path_and_name_and_type: Sequence[Tuple[str, str, str]],
    key_file: Optional[str],
    train_config: Optional[str],
    model_file: Optional[str],
    log_base: Optional[float],
    allow_variable_data_keys: bool,
):
    assert check_argument_types()
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s (%(module)s:%(lineno)d) %(levelname)s: %(message)s",
    )

    if ngpu >= 1:
        device = "cuda"
    else:
        device = "cpu"

    # 1. Set random-seed
    set_all_random_seed(seed)

    # 2. Build LM
    model, train_args = LMTask.build_model_from_file(config_file=train_config, model_file=model_file, device=device)
    # Wrape model to make model.nll() data-parallel
    wrapped_model = ForwardAdaptor(model, "nll")
    wrapped_model.to(dtype=getattr(torch, dtype)).eval()
    logging.info(f"Model:\n{model}")

    # 3. Build data-iterator
    loader = LMTask.build_streaming_iterator(
        data_path_and_name_and_type,
        dtype=dtype,
        batch_size=batch_size,
        key_file=key_file,
        num_workers=num_workers,
        preprocess_fn=LMTask.build_preprocess_fn(train_args, False),
        collate_fn=LMTask.build_collate_fn(train_args, False),
        allow_variable_data_keys=allow_variable_data_keys,
        inference=True,
    )

    # 4. Start for-loop
    with DatadirWriter(output_dir) as writer:
        total_nll = 0.0
        total_ntokens = 0
        for keys, batch in loader:
            assert isinstance(batch, dict), type(batch)
            assert all(isinstance(s, str) for s in keys), keys
            _bs = len(next(iter(batch.values())))
            assert len(keys) == _bs, f"{len(keys)} != {_bs}"

            with torch.no_grad():
                batch = to_device(batch, device)
                if ngpu <= 1:
                    # NOTE(kamo): data_parallel also should work with ngpu=1,
                    # but for debuggability it's better to keep this block.
                    nll, lengths = wrapped_model(**batch)
                else:
                    nll, lengths = data_parallel(
                        wrapped_model, (), range(ngpu), module_kwargs=batch
                    )

            assert _bs == len(nll) == len(lengths), (_bs, len(nll), len(lengths))
            # nll: (B, L) -> (B,)
            nll = nll.detach().cpu().numpy().sum(1)
            # lengths: (B,)
            lengths = lengths.detach().cpu().numpy()
            total_nll += nll.sum()
            total_ntokens += lengths.sum()

            for key, _nll, ntoken in zip(keys, nll, lengths):
                if log_base is None:
                    utt_ppl = np.exp(_nll / ntoken)
                else:
                    utt_ppl = log_base ** (_nll / ntoken / np.log(log_base))

                # Write PPL of each utts for debugging or analysis
                writer["utt2nll"][key] = str(-_nll)
                writer["utt2ppl"][key] = str(utt_ppl)
                writer["utt2ntokens"][key] = str(ntoken)

        if log_base is None:
            ppl = np.exp(total_nll / total_ntokens)
        else:
            ppl = log_base ** (total_nll / total_ntokens / np.log(log_base))

        with (Path(output_dir) / "ppl").open("w", encoding="utf-8") as f:
            f.write(f"{ppl}\n")
        with (Path(output_dir) / "base").open("w", encoding="utf-8") as f:
            if log_base is None:
                _log_base = np.e
            else:
                _log_base = log_base
            f.write(f"{_log_base}\n")
        logging.info(f"PPL={ppl}")


def get_parser():
    parser = config_argparse.ArgumentParser(
        description="Calc perplexity",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # Note(kamo): Use '_' instead of '-' as separator.
    # '-' is confusing if written in yaml.
    parser.add_argument(
        "--log_level",
        type=lambda x: x.upper(),
        default="INFO",
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"),
        help="The verbose level of logging",
    )

    parser.add_argument("--output_dir", type=str, required=True)
    parser.add_argument(
        "--ngpu",
        type=int,
        default=0,
        help="The number of gpus. 0 indicates CPU mode",
    )
    parser.add_argument("--seed", type=int, default=0, help="Random seed")
    parser.add_argument(
        "--dtype",
        default="float32",
        choices=["float16", "float32", "float64"],
        help="Data type",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=1,
        help="The number of workers used for DataLoader",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=1,
        help="The batch size for inference",
    )
    parser.add_argument(
        "--log_base",
        type=float_or_none,
        default=None,
        help="The base of logarithm for Perplexity. "
        "If None, napier's constant is used.",
    )

    group = parser.add_argument_group("Input data related")
    group.add_argument(
        "--data_path_and_name_and_type",
        type=str2triple_str,
        required=True,
        action="append",
    )
    group.add_argument("--key_file", type=str_or_none)
    group.add_argument("--allow_variable_data_keys", type=str2bool, default=False)

    group = parser.add_argument_group("The model configuration related")
    group.add_argument("--train_config", type=str)
    group.add_argument("--model_file", type=str)

    return parser


def main(cmd=None):
    print(get_commandline_args(), file=sys.stderr)
    parser = get_parser()
    args = parser.parse_args(cmd)
    kwargs = vars(args)
    kwargs.pop("config", None)
    calc_perplexity(**kwargs)


if __name__ == "__main__":
    main()



import asyncio
import os
import subprocess
import modal

app = modal.App("seed-x_sglang")

img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvcr.io/nvidia/cuda:12.9.1-cudnn-devel-ubuntu22.04",
        add_python="3.12",
    )
    .pip_install(  # add sglang and some Python dependencies
        # as per sglang website: https://docs.sglang.ai/get_started/install.html
        "flashinfer-python",
        "sglang[all]>=0.5.1.post1",
        extra_options="--find-links https://flashinfer.ai/whl/cu126/torch2.6/flashinfer-python/",
        extra_index_url="https://flashinfer.ai/whl/cu126/torch2.6/",
    )
    .apt_install("libnuma-dev")  # Add NUMA library for sgl_kernel
    .env(
        {
            "TORCH_CUDA_ARCH_LIST": "7.5 8.0 8.6 8.7 8.9 9.0 10.0",
        }
    )
)

img = img.pip_install(
    "achatbot==0.0.23",
    extra_index_url="https://pypi.org/simple/",
)

HF_MODEL_DIR = "/root/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)


with img.imports():
    import uuid
    import torch
    from sglang import Engine
    from sglang.srt.managers.io_struct import GenerateReqInput

    MODEL_PATH = os.getenv("LLM_MODEL", "ByteDance-Seed/Seed-X-PPO-7B")
    model_path = os.path.join(HF_MODEL_DIR, MODEL_PATH)
    model_name = MODEL_PATH.split("/")[-1]


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func, **kwargs):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    subprocess.run("which vllm", shell=True)
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda")
        print(gpu_prop)

    func(**kwargs)


def generate(**kwargs):
    engine = Engine(
        model_path=model_path,
    )

    messages = [
        # without CoT
        "Translate the following English sentence into Chinese:\nMay the force be with you <zh>",
        # with CoT
        "Translate the following English sentence into Chinese and explain it in detail:\nMay the force be with you <zh>",
    ]
    for prompt in messages:
        iterator = engine.generate(  # no request_id param :(
            prompt=prompt,
            sampling_params={
                "n": 1,  # number of samples to generate
                "max_new_tokens": 512,
                "temperature": 0.95,
                "top_p": 1.0,
                "top_k": 20,
                "min_p": 0.0,
                # Penalizers
                "repetition_penalty": 1.1,
                "min_new_tokens": 0,
            },
            stream=True,
            return_logprob=True,
        )
        for part in iterator:
            print(part)
            meta_info = part["meta_info"]
            if "output_token_logprobs" in meta_info and len(meta_info["output_token_logprobs"]) > 0:
                token_id = meta_info["output_token_logprobs"][-1][1]
                print(token_id)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
async def async_generate():
    engine = Engine(
        model_path=model_path,
    )

    messages = [
        # without CoT
        "Translate the following English sentence into Chinese:\nMay the force be with you <zh>",
        # with CoT
        "Translate the following English sentence into Chinese and explain it in detail:\nMay the force be with you <zh>",
    ]
    for prompt in messages:
        obj = GenerateReqInput(
            text=prompt,
            rid=str(uuid.uuid4().hex),
            sampling_params={
                "n": 1,  # number of samples to generate
                "max_new_tokens": 512,
                "temperature": 0.95,
                "top_p": 0.8,
                "top_k": 20,
                "min_p": 0.0,
                # Penalizers
                "repetition_penalty": 1.1,
                "min_new_tokens": 0,
            },
            stream=True,
            return_logprob=True,
        )
        generator = engine.tokenizer_manager.generate_request(obj, None)

        async for part in generator:
            print(part)
            print(part.get("output_ids", []))


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
async def run_achatbot_generator():
    import uuid
    import time

    from transformers import AutoTokenizer, GenerationConfig

    from achatbot.core.llm.sglang.generator import (
        SGlangGenerator,
        SGLangEngineArgs,
        ServerArgs,
    )
    from achatbot.common.session import Session
    from achatbot.common.types import SessionCtx

    generator = SGlangGenerator(
        **SGLangEngineArgs(
            serv_args=ServerArgs(model_path=model_path).__dict__,
        ).__dict__
    )

    tokenizer = AutoTokenizer.from_pretrained(model_path)
    generation_config = {}
    if os.path.exists(os.path.join(model_path, "generation_config.json")):
        generation_config = GenerationConfig.from_pretrained(
            model_path, "generation_config.json"
        ).to_dict()
    # async def run():
    prompt_cases = [
        # without CoT
        {
            "prompt": "Translate the following English sentence into Chinese:\nMay the force be with you <zh>",
            "kwargs": {"max_new_tokens": 30, "stop_ids": []},
        },
        # with CoT
        {
            "prompt": "Translate the following English sentence into Chinese and explain it in detail:\nMay the force be with you <zh>",
            "kwargs": {"max_new_tokens": 512, "stop_ids": []},
        },
    ]

    # test the same session
    # session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    for case in prompt_cases:
        session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
        session.ctx.state["token_ids"] = tokenizer.encode("hello, my name is")
        tokens = tokenizer(case["prompt"])
        session.ctx.state["token_ids"] = tokens["input_ids"]
        gen_kwargs = {**generation_config, **case["kwargs"], **tokens}
        print("gen_kwargs:", gen_kwargs)
        first = True
        start_time = time.perf_counter()
        gen_texts = ""
        async for token_id in generator.generate(session, **gen_kwargs):
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate TTFT time: {ttft} s")
                first = False
            gen_text = tokenizer.decode(token_id)
            gen_texts += gen_text
            print(session.ctx.client_id, token_id, gen_text)
        print(session.ctx.client_id, gen_texts)


"""
IMAGE_GPU=L4 modal run src/llm/sglang/translation/seed_x.py
IMAGE_GPU=L4 modal run src/llm/sglang/translation/seed_x.py::async_generate
IMAGE_GPU=L4 modal run src/llm/sglang/translation/seed_x.py::run_achatbot_generator
"""


@app.local_entrypoint()
def main(
    task: str = "generate",
):
    print(task)
    tasks = {
        "generate": generate,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(
        tasks[task],
    )



import os
import logging
import asyncio
import sys
import uuid
import signal
from typing import Dict, Type, Any, Optional

from apipeline.frames import CancelFrame
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from dotenv import load_dotenv
import nest_asyncio

from achatbot.processors.context.memory import MemoryProcessor
from achatbot.processors.omni.base import VisionVoiceProcessorBase
from achatbot.processors.voice.base import VoiceProcessorBase
from achatbot.processors.image.base import ImageGenProcessor
from achatbot.processors.image import get_image_gen_processor
from achatbot.modules.vision.ocr import VisionOCREnvInit
from achatbot.modules.vision.detector import VisionDetectorEnvInit
from achatbot.processors.ai_processor import AIProcessor, FrameProcessor
from achatbot.processors.vision.vision_processor import MockVisionProcessor
from achatbot.processors.avatar.base import AvatarProcessorBase
from achatbot.processors.speech.asr.base import ASRProcessorBase
from achatbot.processors.llm.base import LLMProcessor
from achatbot.processors.speech.tts.base import TTSProcessorBase
from achatbot.modules.speech.vad_analyzer import VADAnalyzerEnvInit
from achatbot.modules.speech.asr import ASREnvInit
from achatbot.core.llm import LLMEnvInit
from achatbot.modules.speech.tts import TTSEnvInit
from achatbot.modules.avatar import AvatarEnvInit
from achatbot.types.ai_conf import (
    BaseConfig,
    ASRConfig,
    LLMConfig,
    TTSConfig,
    AIConfig,
)
from achatbot.common import interface
from achatbot.common.factory import EngineClass
from achatbot.common.types import BotRunArgs
from achatbot.common.interface import IBot, IVisionDetector
from achatbot.common.session import Session
from achatbot.common.types import SessionCtx
from achatbot.common.logger import Logger
from achatbot.common.pool_engine import EngineProviderPool, PoolInstanceInfo
from achatbot.processors.session_processor import SessionProcessor

load_dotenv(override=True)

use_nest_asyncio = os.getenv("NEST_ASYNCIO", "1") != "0"
if use_nest_asyncio:
    nest_asyncio.apply()


Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)


class AIBot(IBot):
    r"""
    use ai bot config
    !TODONE: need config processor with bot config (redefine api params) @weedge
    bot config: Dict[str, Dict[str,Any]]
    e.g. {"llm":{"key":val,"tag":TAG,"args":{}}, "tts":{"key":val,"tag":TAG,"args":{}}}
    !TIPS: RTVI config options can transfer to ai bot config

    !NOTE:
    use multiprocessing pipe to run bot with unix socket, bot __init__ to new a bot obj must be serializable (pickle); or wraper a func, don't use bot obj method.
    """

    def __init__(self, **args) -> None:
        self.args = BotRunArgs(**args)
        if self.args.bot_name is None or len(self.args.bot_name) == 0:
            self.args.bot_name = self.__class__.__name__

        self.task: PipelineTask | None = None
        self.session = Session(**SessionCtx(str(uuid.uuid4())).__dict__)
        self.runner: PipelineRunner | None = None
        self.generator: interface.ILlmGenerator = None

        self._bot_config_list = self.args.bot_config_list
        self._bot_config: AIConfig | None = None
        self._handle_sigint = self.args.handle_sigint
        self._save_audio = self.args.save_audio

        self._pipeline_params = PipelineParams()
        if self.args.pipeline:
            self._pipeline_params = PipelineParams(**self.args.pipeline)

        self.vad_analyzer_pool: EngineProviderPool = None
        self.asr_pool: EngineProviderPool = None
        self.translate_llm_generator_pool: EngineProviderPool = None
        self.tts_pool: EngineProviderPool = None
        self.avatar_pool: EngineProviderPool = None

    def init_bot_config(self):
        try:
            logging.debug(f"args.bot_config: {self.args.bot_config}")
            self._bot_config: AIConfig = AIConfig(**self.args.bot_config)
            if len(self._bot_config_list) > 0:
                from achatbot.types.rtvi import RTVIConfig

                rtvi_config = RTVIConfig(config_list=self.args.bot_config_list)
                self._bot_config = AIConfig(**rtvi_config._arguments_dict)
            if self._bot_config.llm is None:
                self._bot_config.llm = LLMConfig()
        except Exception as e:
            raise Exception(f"Failed to parse bot configuration: {e}")
        logging.info(f"ai bot_config: {self._bot_config}")

    def set_args(self, args):
        merge_args = {**self.args.__dict__, **args}
        self.args = BotRunArgs(**merge_args)

    def bot_config(self):
        return self._bot_config

    def load(self):
        """
        NOTE:
        - init modules here
            # load model ckpt when bot start
            # when deploy need load model ckpt, then run serve
        - don't init processor; processor for each connect session
        - load model engine here to share,
        - if don't need share, init in processor, u can init a engine pool
        have fun :)
        """
        pass

    def run(self):
        asyncio.run(self.async_run())

    def __del__(self):
        self.cancel()

    def cancel(self):
        if self.runner:
            asyncio.create_task(self.runner.cancel())
        if self.generator:
            self.generator.close()

    async def async_run(self):
        try:
            await self.arun()
        except asyncio.CancelledError:
            logging.info("CancelledError, Exiting!")
            if self.task is not None:
                logging.info("Cancelling task!")
                await self.task.queue_frame(CancelFrame())
        except KeyboardInterrupt:
            logging.warning("Ctrl-C detected. Exiting!")
            if self.task is not None:
                logging.info("Cancelling task!")
                await self.task.queue_frame(CancelFrame())
        except Exception as e:
            logging.error(f"run error: {e}", exc_info=True)
        finally:
            if self.args.handle_sigint is True:
                loop = asyncio.get_running_loop()
                loop.remove_signal_handler(signal.SIGINT)
                loop.remove_signal_handler(signal.SIGTERM)
            self.cancel()
            logging.info(f"{__name__} task run is Done!")

    async def arun(self):
        pass

    def get_pool(self, conf: BaseConfig, creation_func: callable) -> EngineProviderPool | None:
        pool: EngineProviderPool = None
        if conf and conf.pool_size:
            pool = EngineProviderPool(
                pool_size=conf.pool_size,
                new_func=creation_func,
                init_worker_num=conf.pool_init_worker_num,
            )

        if pool and not pool.initialize():
            logging.error(f"Failed to initialize pool func:{creation_func} with conf:{conf}")
            return None

        return pool

    def get_vad_analyzer_pool(self):
        return self.get_pool(self._bot_config.vad, self.get_vad_analyzer)

    def get_vad_analyzer_from_pool(
        self,
    ) -> tuple[Optional[PoolInstanceInfo], interface.IVADAnalyzer | EngineClass]:
        vad_analyzer_info = None
        if self.vad_analyzer_pool:
            vad_analyzer_info = self.vad_analyzer_pool.get()
        vad_analyzer = (
            vad_analyzer_info.get_instance() if vad_analyzer_info else self.get_vad_analyzer()
        )
        return (vad_analyzer_info, vad_analyzer)

    def get_vad_analyzer(self) -> interface.IVADAnalyzer | EngineClass:
        vad_analyzer: interface.IVADAnalyzer | EngineClass = None
        if (
            self._bot_config.vad
            and self._bot_config.vad.tag
            and len(self._bot_config.vad.tag) > 0
            and self._bot_config.vad.args
        ):
            vad_analyzer = VADAnalyzerEnvInit.getEngine(
                self._bot_config.vad.tag, **self._bot_config.vad.args
            )
        else:
            vad_analyzer = VADAnalyzerEnvInit.initVADAnalyzerEngine()
        return vad_analyzer

    def get_avatar_pool(self):
        return self.get_pool(self._bot_config.avatar, self.get_avatar)

    def get_avatar_from_pool(self) -> tuple[Optional[PoolInstanceInfo], EngineClass | None]:
        avatar_info = None
        if self.avatar_pool:
            avatar_info = self.avatar_pool.get()
        avatar = avatar_info.get_instance() if avatar_info else self.get_avatar()
        return (avatar_info, avatar)

    def get_avatar(self) -> EngineClass | None:
        avatar: EngineClass | None = None
        if self._bot_config.avatar and self._bot_config.avatar.tag:
            if self._bot_config.avatar.tag == "lam_audio2expression_avatar":
                from achatbot.modules.avatar.lam_audio2expression import LAMAudio2ExpressionAvatar

                if self._bot_config.avatar.args:
                    avatar = LAMAudio2ExpressionAvatar(**self._bot_config.avatar.args)
                else:
                    avatar = LAMAudio2ExpressionAvatar()
            else:
                # TODO: use avatar engine
                args = self._bot_config.avatar.args or {}
                avatar = AvatarEnvInit.getEngine(self._bot_config.avatar.tag, **args)
                raise NotImplementedError("don't support use tag to create avatar engine")
            avatar.load()
            return avatar
        else:
            from achatbot.modules.avatar.lite_avatar import LiteAvatar

            logging.info("use default lite avatar engine processor")
            if self._bot_config.avatar and self._bot_config.avatar.args:
                avatar = LiteAvatar(**self._bot_config.avatar.args)
            else:
                avatar = LiteAvatar()
            avatar.load()
        return avatar

    def get_a2a_processor(self, tag: str | None = None) -> SessionProcessor:
        processor: SessionProcessor | None = None
        if self._bot_config.a2a and self._bot_config.a2a.tag and self._bot_config.a2a.args:
            tag = tag or self._bot_config.a2a.tag
            if tag == "a2a_conversation_processor":
                from achatbot.processors.a2a.a2a_conversation_processor import A2AConversationProcessor

                processor = A2AConversationProcessor(
                    session=self.session, **self._bot_config.a2a.args
                )
                return processor
            if tag == "a2a_live_processor":
                from achatbot.processors.a2a.a2a_live_processor import A2ALiveProcessor

                processor = A2ALiveProcessor(session=self.session, **self._bot_config.a2a.args)
                return processor

        raise ValueError("A2A processor is not configured. Please check your bot configuration.")

    def get_avatar_processor(self, avatar=None) -> AvatarProcessorBase:
        avatar_processor: AvatarProcessorBase | None = None
        # use avatar engine processor

        if self._bot_config.avatar and self._bot_config.avatar.tag:
            if self._bot_config.avatar.tag == "lam_audio2expression_avatar":
                from achatbot.processors.avatar.lam_audio2expression_avatar_processor import (
                    LAMAudio2ExpressionAvatarProcessor,
                )
                from achatbot.modules.avatar.lam_audio2expression import LAMAudio2ExpressionAvatar

                if self._bot_config.avatar and self._bot_config.avatar.args:
                    avatar = avatar or LAMAudio2ExpressionAvatar(**self._bot_config.avatar.args)
                else:
                    avatar = avatar or LAMAudio2ExpressionAvatar()
                return LAMAudio2ExpressionAvatarProcessor(avatar, **self._bot_config.avatar.args)
            else:
                # TODO: use avatar engine processor
                args = self._bot_config.avatar.args or {}
                avatar = AvatarEnvInit.getEngine(self._bot_config.avatar.tag, **args)
                _ = avatar
                raise NotImplementedError("don't support use tag to create avatar engine")
                return avatar_processor
        else:
            from achatbot.processors.avatar.lite_avatar_processor import LiteAvatarProcessor
            from achatbot.modules.avatar.lite_avatar import LiteAvatar

            logging.info("use default lite avatar engine processor")
            if self._bot_config.avatar and self._bot_config.avatar.args:
                avatar = avatar or LiteAvatar(**self._bot_config.avatar.args)
            else:
                avatar = avatar or LiteAvatar()
            return LiteAvatarProcessor(avatar)

    def get_asr_pool(self):
        return self.get_pool(self._bot_config.asr, self.get_asr)

    def get_asr_from_pool(
        self,
    ) -> tuple[Optional[PoolInstanceInfo], interface.IAsr | EngineClass | None]:
        asr_info = None
        if self.asr_pool:
            asr_info = self.asr_pool.get()
        asr = asr_info.get_instance() if asr_info else self.get_asr()
        return (asr_info, asr)

    def get_asr(self) -> interface.IAsr | EngineClass | None:
        asr: interface.IAsr | EngineClass | None = None
        if (
            self._bot_config.asr
            and self._bot_config.asr.tag
            and self._bot_config.asr.tag == "deepgram_asr_processor"
            and self._bot_config.asr.args
        ):
            pass
        else:
            if self._bot_config.asr and self._bot_config.asr.tag and self._bot_config.asr.args:
                asr = ASREnvInit.getEngine(self._bot_config.asr.tag, **self._bot_config.asr.args)
            else:
                logging.info("use default asr engine processor")
                asr = ASREnvInit.initASREngine()
        return asr

    def get_asr_processor(
        self, asr_engine: interface.IAsr | EngineClass | None = None
    ) -> ASRProcessorBase:
        asr_processor: ASRProcessorBase | None = None
        if (
            self._bot_config.asr
            and self._bot_config.asr.tag
            and self._bot_config.asr.tag == "deepgram_asr_processor"
            and self._bot_config.asr.args
        ):
            from achatbot.processors.speech.asr.deepgram_asr_processor import DeepgramAsrProcessor

            asr_processor = DeepgramAsrProcessor(
                api_key=os.getenv("DEEPGRAM_API_KEY"), **self._bot_config.asr.args
            )
        else:
            # use asr engine processor
            from achatbot.processors.speech.asr.asr_processor import ASRProcessor

            if self._bot_config.asr and self._bot_config.asr.tag and self._bot_config.asr.args:
                asr = asr_engine or ASREnvInit.getEngine(
                    self._bot_config.asr.tag, **self._bot_config.asr.args
                )
            else:
                logging.info("use default asr engine processor")
                asr = asr_engine or ASREnvInit.initASREngine()
                self._bot_config.asr = ASRConfig(tag=asr.SELECTED_TAG, args=asr.get_args_dict())
            asr_processor = ASRProcessor(asr=asr, session=self.session)
        return asr_processor

    def get_hf_tokenizer(self):
        """
        To disable this warning, you can either:
           - Avoid using `tokenizers` before the fork if possible
           - Explicitly set the environment variable TOKENIZERS_PARALLELISM=(true | false)
        """
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained(self._bot_config.translate_llm.model)
        return tokenizer

    def get_translate_llm_generator_pool(self):
        return self.get_pool(self._bot_config.translate_llm, self.get_translate_llm_generator)

    def get_translate_llm_generator_from_pool(
        self,
    ) -> tuple[
        Optional[PoolInstanceInfo], interface.ILlmGenerator | interface.ILlm | EngineClass | None
    ]:
        translate_llm_generator_info = None
        if self.translate_llm_generator_pool:
            translate_llm_generator_info = self.translate_llm_generator_pool.get()
        translate_llm_generator = (
            translate_llm_generator_info.get_instance()
            if translate_llm_generator_info
            else self.get_translate_llm_generator()
        )
        return (translate_llm_generator_info, translate_llm_generator)

    def get_translate_llm_generator(self):
        """
        get local translate llm generator
        """
        assert self._bot_config.translate_llm, f"translate_llm must be provided {self._bot_config=}"
        # load llm generator
        tag = self._bot_config.translate_llm.tag
        args = self._bot_config.translate_llm.args or {}
        generator = LLMEnvInit.initLLMEngine(
            tag=tag,
            kwargs=args,
        )
        return generator

    def get_vision_llm_processor(
        self, llm_config: LLMConfig | None = None, llm_engine=None
    ) -> LLMProcessor:
        """
        get local vision llm
        """
        from achatbot.processors.vision.vision_processor import VisionProcessor

        if not llm_config:
            llm_config = self._bot_config.vision_llm
        if "mock" in llm_config.tag:
            llm_processor = MockVisionProcessor()
        else:
            logging.info(f"init engine llm processor tag: {llm_config.tag}")
            sleep_time_s = llm_config.args.pop("sleep_time_s", 0.15)
            llm_engine = llm_engine or LLMEnvInit.initLLMEngine(llm_config.tag, llm_config.args)
            llm_processor = VisionProcessor(llm_engine, self.session, sleep_time_s=sleep_time_s)
        return llm_processor

    def get_remote_llm_processor(self, llm: LLMConfig | None = None) -> LLMProcessor:
        """
        get remote llm
        """
        if not llm:
            llm = self._bot_config.llm
        if llm and llm.tag and "google" in llm.tag:
            llm_processor = self.get_google_llm_processor(llm)
        elif llm and llm.tag and "litellm" in llm.tag:
            llm_processor = self.get_litellm_processor(llm)
        elif llm and llm.tag and "openai" in llm.tag:
            llm_processor = self.get_openai_llm_processor(llm)
        else:
            llm_processor = self.get_google_llm_processor(llm)
        return llm_processor

    def get_openai_llm_processor(self, llm: LLMConfig | None = None) -> LLMProcessor:
        from achatbot.processors.llm.openai_llm_processor import (
            OpenAILLMProcessor,
            OpenAIGroqLLMProcessor,
        )

        if not llm:
            llm = self._bot_config.llm
        # default use openai llm processor
        api_key = os.environ.get("OPENAI_API_KEY")
        if llm:
            if "groq" in llm.base_url:
                # https://console.groq.com/docs/models
                api_key = os.environ.get("GROQ_API_KEY")
                llm_processor = OpenAIGroqLLMProcessor(
                    model=llm.model,
                    base_url=llm.base_url,
                    api_key=api_key,
                )
                return llm_processor
            elif "together" in llm.base_url:
                # https://docs.together.ai/docs/chat-models
                api_key = os.environ.get("TOGETHER_API_KEY")
            elif "local" in llm.base_url or "127.0.0.1" in llm.base_url:
                api_key = "ollama"

        llm_processor = OpenAILLMProcessor(
            model=llm.model,
            base_url=llm.base_url,
            api_key=api_key,
        )
        return llm_processor

    def get_google_llm_processor(self, llm: LLMConfig) -> LLMProcessor:
        from achatbot.processors.llm.google_llm_processor import GoogleAILLMProcessor

        llm_config = llm
        if llm_config and llm_config.args:
            logging.info(f"use google llm processor args:{llm_config.args}")
            llm_processor = GoogleAILLMProcessor(**llm_config.args)
        else:
            logging.info("use default google llm processor")
            api_key = os.environ.get("GOOGLE_API_KEY")
            model = os.environ.get("GOOGLE_LLM_MODEL", "gemini-1.5-flash-latest")
            llm_processor = GoogleAILLMProcessor(
                api_key=api_key,
                model=model,
            )

        return llm_processor

    def get_litellm_processor(self, llm: LLMConfig) -> LLMProcessor:
        from achatbot.processors.llm.litellm_processor import LiteLLMProcessor

        llm_processor = LiteLLMProcessor(model=llm.model, set_verbose=False)
        return llm_processor

    def get_llm_processor(self, llm: LLMConfig | None = None, llm_engine=None) -> LLMProcessor:
        if not llm:
            llm = self._bot_config.llm or self._bot_config.vision_llm
        if llm and llm.tag and "vision" in llm.tag:
            # engine llm processor(just support vision model, other TODO):
            # (llm_llamacpp, llm_personalai_proxy, llm_transformers etc..)
            llm_processor = self.get_vision_llm_processor(llm, llm_engine)
        else:
            llm_processor = self.get_remote_llm_processor(llm)
        return llm_processor

    def get_vision_annotate_processor(self) -> AIProcessor:
        from achatbot.processors.vision.annotate_processor import AnnotateProcessor

        detector: IVisionDetector | EngineClass = VisionDetectorEnvInit.initVisionDetectorEngine(
            self._bot_config.vision_detector.tag, self._bot_config.vision_detector.args
        )
        processor = AnnotateProcessor(detector, self.session)
        return processor

    def get_vision_detect_processor(self) -> AIProcessor:
        from achatbot.processors.vision.detect_processor import DetectProcessor

        detector = VisionDetectorEnvInit.initVisionDetectorEngine(
            self._bot_config.vision_detector.tag, self._bot_config.vision_detector.args
        )

        desc, out_desc = "", ""
        if "desc" in self._bot_config.vision_detector.args:
            desc = self._bot_config.vision_detector.args["desc"]
        if "out_desc" in self._bot_config.vision_detector.args:
            out_desc = self._bot_config.vision_detector.args["out_desc"]
        if desc and out_desc:
            processor = DetectProcessor(
                detected_text=desc,
                out_detected_text=out_desc,
                detector=detector,
                session=self.session,
            )
        elif desc:
            processor = DetectProcessor(detected_text=desc, detector=detector, session=self.session)
        elif out_desc:
            processor = DetectProcessor(
                out_detected_text=out_desc, detector=detector, session=self.session
            )
        else:
            processor = DetectProcessor(detector=detector, session=self.session)

        return processor

    def get_vision_ocr_processor(self) -> AIProcessor:
        from achatbot.processors.vision.ocr_processor import OCRProcessor

        ocr = VisionOCREnvInit.initVisionOCREngine(
            self._bot_config.vision_ocr.tag, self._bot_config.vision_ocr.args
        )
        processor = OCRProcessor(ocr=ocr, session=self.session)
        return processor

    def get_moshi_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        if not llm:
            llm = self._bot_config.voice_llm
        if llm and llm.tag and "moshi" in llm.tag:
            from achatbot.processors.voice.moshi_voice_processor import (
                MoshiVoiceOpusStreamProcessor,
                MoshiVoiceProcessor,
            )

            if "moshi_opus" in llm.tag:
                if llm.args:
                    llm_processor = MoshiVoiceOpusStreamProcessor(**llm.args)
                else:
                    llm_processor = MoshiVoiceOpusStreamProcessor()
            else:
                if llm.args:
                    llm_processor = MoshiVoiceProcessor(**llm.args)
                else:
                    llm_processor = MoshiVoiceProcessor()
        else:
            from achatbot.processors.voice.moshi_voice_processor import VoiceOpusStreamEchoProcessor

            llm_processor = VoiceOpusStreamEchoProcessor()
        return llm_processor

    def get_text_minicpmo_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.minicpmo_voice_processor import MiniCPMoTextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = MiniCPMoTextVoiceProcessor(**llm.args)
        else:
            llm_processor = MiniCPMoTextVoiceProcessor()
        return llm_processor

    def get_text_qwen2_5omni_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VoiceProcessorBase:
        from achatbot.processors.voice.qwen2_5omni_voice_processor import Qwen2_5OmniTextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = Qwen2_5OmniTextVoiceProcessor(**llm.args)
        else:
            llm_processor = Qwen2_5OmniTextVoiceProcessor()
        return llm_processor

    def get_text_kimi_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.kimi_voice_processor import KimiTextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = KimiTextVoiceProcessor(**llm.args)
        else:
            llm_processor = KimiTextVoiceProcessor()
        return llm_processor

    def get_text_vita_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.vita_voice_processor import VITATextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = VITATextVoiceProcessor(**llm.args)
        else:
            llm_processor = VITATextVoiceProcessor()
        return llm_processor

    def get_audio_minicpmo_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VoiceProcessorBase:
        from achatbot.processors.voice.minicpmo_voice_processor import MiniCPMoAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = MiniCPMoAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = MiniCPMoAudioVoiceProcessor()
        return llm_processor

    def get_audio_qwen2_5omni_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VoiceProcessorBase:
        from achatbot.processors.voice.qwen2_5omni_voice_processor import Qwen2_5OmniAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = Qwen2_5OmniAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = Qwen2_5OmniAudioVoiceProcessor()
        return llm_processor

    def get_audio_kimi_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.kimi_voice_processor import KimiAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = KimiAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = KimiAudioVoiceProcessor()
        return llm_processor

    def get_audio_vita_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.vita_voice_processor import VITAAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = VITAAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = VITAAudioVoiceProcessor()
        return llm_processor

    def get_audio_phi4_speech_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.phi4_speech_processor import Phi4AudioTextProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = Phi4AudioTextProcessor(**llm.args)
        else:
            llm_processor = Phi4AudioTextProcessor()
        return llm_processor

    def get_minicpmo_vision_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VisionVoiceProcessorBase:
        from achatbot.processors.omni.minicpmo_vision_voice import MiniCPMoVisionVoiceProcessor
        from achatbot.processors.omni.base import MockVisionVoiceProcessor

        if not llm:
            llm = self._bot_config.omni_llm
        if "mock" in llm.tag:
            return MockVisionVoiceProcessor()
        if llm.args:
            llm_processor = MiniCPMoVisionVoiceProcessor(**llm.args)
        else:
            llm_processor = MiniCPMoVisionVoiceProcessor()
        return llm_processor

    def get_qwen2_5omni_vision_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VisionVoiceProcessorBase:
        from achatbot.processors.omni.qwen2_5omni_vision_voice import Qwen2_5OmnVisionVoiceProcessor
        from achatbot.processors.omni.base import MockVisionVoiceProcessor

        if not llm:
            llm = self._bot_config.omni_llm
        if "mock" in llm.tag:
            return MockVisionVoiceProcessor()
        if llm.args:
            llm_processor = Qwen2_5OmnVisionVoiceProcessor(**llm.args)
        else:
            llm_processor = Qwen2_5OmnVisionVoiceProcessor()
        return llm_processor

    def get_phi4_vision_speech_processor(
        self, llm: LLMConfig | None = None
    ) -> VisionVoiceProcessorBase:
        from achatbot.processors.omni.phi4_vision_speech import Phi4VisionSpeechProcessor
        from achatbot.processors.omni.base import MockVisionVoiceProcessor

        if not llm:
            llm = self._bot_config.omni_llm
        if "mock" in llm.tag:
            return MockVisionVoiceProcessor()
        if llm.args:
            llm_processor = Phi4VisionSpeechProcessor(**llm.args)
        else:
            llm_processor = Phi4VisionSpeechProcessor()
        return llm_processor

    def get_gemma3n_vision_speech_processor(
        self, llm: LLMConfig | None = None
    ) -> VisionVoiceProcessorBase:
        from achatbot.processors.omni.gemma_vision_speech import Gemma3nVisionSpeechProcessor
        from achatbot.processors.omni.base import MockVisionVoiceProcessor

        if not llm:
            llm = self._bot_config.omni_llm
        if "mock" in llm.tag:
            return MockVisionVoiceProcessor()
        if llm.args:
            llm_processor = Gemma3nVisionSpeechProcessor(**llm.args)
        else:
            llm_processor = Gemma3nVisionSpeechProcessor()
        return llm_processor

    def get_text_glm_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.glm_voice_processor import GLMTextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = GLMTextVoiceProcessor(**llm.args)
        else:
            llm_processor = GLMTextVoiceProcessor()
        return llm_processor

    def get_audio_glm_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.glm_voice_processor import GLMAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = GLMAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = GLMAudioVoiceProcessor()
        return llm_processor

    def get_text_step_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.step_voice_processor import (
            StepTextVoiceProcessor,
            MockStepTextVoiceProcessor,
        )

        if not llm:
            llm = self._bot_config.voice_llm
        if "mock" in llm.tag:
            return MockStepTextVoiceProcessor()
        if llm.args:
            llm_processor = StepTextVoiceProcessor(**llm.args)
        else:
            llm_processor = StepTextVoiceProcessor()
        return llm_processor

    def get_audio_step_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from achatbot.processors.voice.step_voice_processor import (
            StepAudioVoiceProcessor,
            MockStepAudioVoiceProcessor,
        )

        if not llm:
            llm = self._bot_config.voice_llm
        if "mock" in llm.tag:
            return MockStepAudioVoiceProcessor()
        if llm.args:
            llm_processor = StepAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = StepAudioVoiceProcessor()
        return llm_processor

    def get_audio_freeze_omni_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VoiceProcessorBase:
        cur_dir = os.path.dirname(__file__)
        if bool(os.getenv("ACHATBOT_PKG", "")):
            sys.path.insert(1, os.path.join(cur_dir, "../../FreezeOmni"))
        else:
            sys.path.insert(1, os.path.join(cur_dir, "../../../deps/FreezeOmni"))

        from achatbot.processors.voice.freeze_omni_voice_processor import FreezeOmniVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        # TODO: need create inference_pipeline_pool and tts pool
        if llm.args:
            llm_processor = FreezeOmniVoiceProcessor(**llm.args)
        else:
            llm_processor = FreezeOmniVoiceProcessor()
        return llm_processor

    def get_openai_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        # TODO: use openai reatime voice api
        pass

    def get_tts_pool(self):
        return self.get_pool(self._bot_config.tts, self.get_tts)

    def get_tts_from_pool(
        self,
    ) -> tuple[Optional[PoolInstanceInfo], interface.ITts | EngineClass | None]:
        tts_info = None
        if self.tts_pool:
            tts_info = self.tts_pool.get()
        tts = tts_info.get_instance() if tts_info else self.get_tts()
        return (tts_info, tts)

    def get_tts(self) -> interface.ITts | EngineClass | None:
        tts: interface.ITts | EngineClass | None = None
        if (
            self._bot_config.tts
            and self._bot_config.tts.tag
            and self._bot_config.tts.tag in ["elevenlabs_tts_processor", "cartesia_tts_processor"]
        ):
            pass
        else:
            if self._bot_config.tts and self._bot_config.tts.tag and self._bot_config.tts.args:
                tts = TTSEnvInit.getEngine(self._bot_config.tts.tag, **self._bot_config.tts.args)
            else:
                logging.info("use default tts engine processor")
                tts = TTSEnvInit.initttsEngine()
        return tts

    def get_tts_processor(
        self, tts_engine: interface.ITts | EngineClass = None
    ) -> TTSProcessorBase:
        tts_processor: TTSProcessorBase | None = None
        if self._bot_config.tts and self._bot_config.tts.tag and self._bot_config.tts.args:
            if self._bot_config.tts.tag == "elevenlabs_tts_processor":
                from achatbot.processors.speech.tts.elevenlabs_tts_processor import (
                    ElevenLabsTTSProcessor,
                )

                tts_processor = ElevenLabsTTSProcessor(**self._bot_config.tts.args)
            elif self._bot_config.tts.tag == "cartesia_tts_processor":
                from achatbot.processors.speech.tts.cartesia_tts_processor import CartesiaTTSProcessor

                tts_processor = CartesiaTTSProcessor(**self._bot_config.tts.args)
            else:
                # use tts engine processor
                from achatbot.processors.speech.tts.tts_processor import TTSProcessor

                tts_engine = tts_engine or TTSEnvInit.getEngine(
                    self._bot_config.tts.tag, **self._bot_config.tts.args
                )
                self._bot_config.tts.tag = tts_engine.SELECTED_TAG
                self._bot_config.tts.args = tts_engine.get_args_dict()
                tts_processor = TTSProcessor(
                    tts=tts_engine,
                    session=self.session,
                    aggregate_sentences=self._bot_config.tts.aggregate_sentences,
                    push_text_frames=self._bot_config.tts.push_text_frames,
                    remove_punctuation=self._bot_config.tts.remove_punctuation,
                )
        else:
            # default tts engine processor
            from achatbot.processors.speech.tts.tts_processor import TTSProcessor

            logging.info("use default tts engine processor")
            tag = None
            if self._bot_config.tts and self._bot_config.tts.tag:
                tag = self._bot_config.tts.tag
            tts_engine = tts_engine or TTSEnvInit.initTTSEngine(tag)
            self._bot_config.tts = TTSConfig(
                tag=tts_engine.SELECTED_TAG, args=tts_engine.get_args_dict()
            )
            tts_processor = TTSProcessor(
                tts=tts_engine,
                session=self.session,
                aggregate_sentences=self._bot_config.tts.aggregate_sentences,
                push_text_frames=self._bot_config.tts.push_text_frames,
                remove_punctuation=self._bot_config.tts.remove_punctuation,
            )

        return tts_processor

    def get_image_gen_processor(self) -> ImageGenProcessor:
        if not self._bot_config.img_gen or not self._bot_config.img_gen.args:
            raise Exception("need img_gen args params")
        return get_image_gen_processor(
            self._bot_config.img_gen.tag, **self._bot_config.img_gen.args
        )

    def get_memory_processor(
        self,
        local_config: Optional[Dict[str, Any]] = None,
    ) -> MemoryProcessor:
        if not self._bot_config.memory or not self._bot_config.memory.processor:
            raise ValueError("Memory processor configuration is missing or invalid.")

        processor_type = self._bot_config.memory.processor
        args = self._bot_config.memory.args or {}

        if processor_type == "Mem0MemoryProcessor":
            from achatbot.processors.context.memory.mem0 import Mem0MemoryProcessor

            return Mem0MemoryProcessor(local_config=local_config, **args)

        raise NotImplementedError(f"Memory processor '{processor_type}' is not supported.")


class AIRoomBot(AIBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)


class AIChannelBot(AIBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)



import collections.abc
from pathlib import Path
from typing import Union

import numpy as np
from typeguard import check_argument_types

from funasr_local.fileio.read_text import read_2column_text


class NpyScpWriter:
    """Writer class for a scp file of numpy file.

    Examples:
        key1 /some/path/a.npy
        key2 /some/path/b.npy
        key3 /some/path/c.npy
        key4 /some/path/d.npy
        ...

        >>> writer = NpyScpWriter('./data/', './data/feat.scp')
        >>> writer['aa'] = numpy_array
        >>> writer['bb'] = numpy_array

    """

    def __init__(self, outdir: Union[Path, str], scpfile: Union[Path, str]):
        assert check_argument_types()
        self.dir = Path(outdir)
        self.dir.mkdir(parents=True, exist_ok=True)
        scpfile = Path(scpfile)
        scpfile.parent.mkdir(parents=True, exist_ok=True)
        self.fscp = scpfile.open("w", encoding="utf-8")

        self.data = {}

    def get_path(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        assert isinstance(value, np.ndarray), type(value)
        p = self.dir / f"{key}.npy"
        p.parent.mkdir(parents=True, exist_ok=True)
        np.save(str(p), value)
        self.fscp.write(f"{key} {p}\n")

        # Store the file path
        self.data[key] = str(p)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.fscp.close()


class NpyScpReader(collections.abc.Mapping):
    """Reader class for a scp file of numpy file.

    Examples:
        key1 /some/path/a.npy
        key2 /some/path/b.npy
        key3 /some/path/c.npy
        key4 /some/path/d.npy
        ...

        >>> reader = NpyScpReader('npy.scp')
        >>> array = reader['key1']

    """

    def __init__(self, fname: Union[Path, str]):
        assert check_argument_types()
        self.fname = Path(fname)
        self.data = read_2column_text(fname)

    def get_path(self, key):
        return self.data[key]

    def __getitem__(self, key) -> np.ndarray:
        p = self.data[key]
        return np.load(p)

    def __contains__(self, item):
        return item

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def keys(self):
        return self.data.keys()



import os
import logging
import time
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, Type, Any, Optional
import queue

from src.common.interface import IPoolInstance
from src.common.register import Register


# 用于存储创建引擎函数的全局字典 use engine factory to create with bot config
# _new_func_map: Dict[str, callable] = {}


class PoolInstanceInfo:
    """池中对象实例的包装信息，包含使用状态等。"""

    def __init__(self, instance: IPoolInstance):
        self.instance_id = None  # 由外部设置
        self._is_in_use = False
        self.last_used = time.time_ns()
        self.instance = instance
        self._lock = threading.Lock()

    def get_instance(self) -> IPoolInstance:
        return self.instance

    def mark_in_use(self) -> bool:
        """尝试标记实例为正在使用。"""
        with self._lock:
            if not self._is_in_use:
                self._is_in_use = True
                self.last_used = time.time_ns()
                return True
            return False

    def mark_available(self) -> bool:
        """尝试标记实例为可用。"""
        with self._lock:
            if self._is_in_use:
                self._is_in_use = False
                self.last_used = time.time_ns()
                return True
            return False

    def is_in_use(self) -> bool:
        with self._lock:
            return self._is_in_use


class EngineProviderPool:
    """engine provider pool (EngineClass)"""

    def __init__(self, pool_size: int, new_func: callable, init_worker_num: int | None = None):
        if pool_size <= 0:
            pool_size = 1
        if init_worker_num and init_worker_num > pool_size:
            init_worker_num = pool_size
        self._pool_instances: queue.Queue[PoolInstanceInfo] = queue.Queue(maxsize=pool_size)
        self.pool_size = pool_size
        self.init_worker_num = init_worker_num or os.cpu_count()  # logic cpu count
        self._new_func = new_func
        self._stop_event = threading.Event()
        self._lock = threading.RLock()  # 用于保护共享计数器

        # 统计信息
        self._total_created = 0
        self._total_reused = 0
        self._total_active = 0

        if not self._new_func:
            raise ValueError(f"No NewFunc registered for {new_func=}")

    def _create_new_instance_info(self) -> PoolInstanceInfo:
        """创建新的实例信息对象。"""
        try:
            instance = self._new_func()
            if not isinstance(instance, IPoolInstance):
                raise TypeError(f"NewFunc for {self._new_func} did not return an IPoolInstance")
        except Exception as e:
            raise RuntimeError(f"Failed to create new instance for {self._new_func}: {e}")

        with self._lock:
            instance_id = self._total_created + 1
            self._total_created = instance_id

        instance_info = PoolInstanceInfo(instance)
        instance_info.instance_id = instance_id
        logging.info(f"Created New InstanceInfo# {instance_info.instance_id}")

        return instance_info

    def initialize(self) -> bool:
        """并行初始化池中的对象。"""
        logging.info(
            f"Initializing pool with {self.init_worker_num=} {self.pool_size} instances..."
        )
        success_count = 0
        init_errors = []

        # 使用 ThreadPoolExecutor 来并行创建实例
        with ThreadPoolExecutor(max_workers=self.init_worker_num) as executor:
            futures = [
                executor.submit(self._create_new_instance_info) for _ in range(self.pool_size)
            ]

            for i, future in enumerate(futures):
                try:
                    instance_info = future.result()
                    try:
                        self._pool_instances.put_nowait(instance_info)
                        logging.info(f"{self._new_func} Instance#{i} Initialized and put into pool")
                        success_count += 1
                    except queue.Full:
                        # 如果队列满了，释放实例
                        instance_info.instance.release()
                        init_errors.append(
                            RuntimeError(
                                f"Pool queue full during initialization, instance#{i} released"
                            )
                        )
                        logging.warning(
                            f"Pool queue full during initialization, instance#{i} released"
                        )
                except Exception as e:
                    init_errors.append(e)
                    logging.warning(f"Initialization warning for instance#{i}: {e}")

        logging.info(
            f"Pool initialized with {success_count}/{self.pool_size} {self._new_func} instances"
        )

        if init_errors and success_count == 0:
            logging.error(f"Failed to initialize any {self._new_func} instances")
            return False

        return True

    def get(self) -> Optional[PoolInstanceInfo]:
        """从池中获取一个实例。"""
        logging.info(
            f"Attempting to get {self._new_func} instance from pool (available: {self._pool_instances.qsize()})"
        )

        # 首先尝试从池中获取
        try:
            instance_info = self._pool_instances.get_nowait()
            if instance_info.mark_in_use():
                with self._lock:
                    self._total_reused += 1
                    self._total_active += 1
                logging.info(
                    f"Got {self._new_func} instanceInfo# {instance_info.instance_id} from pool and marked as in-use (active: {self._total_active})"
                )
                return instance_info
            else:
                # 如果获取到的实例正在被使用，理论上不应该发生，但为了安全，放回队列
                logging.warning(
                    f"Instance# {instance_info.instance_id} was marked as in-use immediately after being taken from pool. This is unexpected."
                )
                try:
                    self._pool_instances.put_nowait(instance_info)
                except queue.Full:
                    # 如果放不回去，释放它
                    logging.warning(
                        f"Could not return unexpected in-use instance# {instance_info.instance_id} to pool, releasing it."
                    )
                    instance_info.instance.release()
        except queue.Empty:
            pass  # 池中没有可用实例

        # 如果池中没有可用实例，等待一段时间看是否有实例被归还
        timeout = 0.1  # 100ms
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                instance_info = self._pool_instances.get(
                    timeout=timeout - (time.time() - start_time)
                )
                if instance_info.mark_in_use():
                    with self._lock:
                        self._total_reused += 1
                        self._total_active += 1
                    logging.info(
                        f"Got {self._new_func} instanceInfo# {instance_info.instance_id} from pool after waiting and marked as in-use (active: {self._total_active})"
                    )
                    return instance_info
                else:
                    # 实例被占用，尝试放回
                    try:
                        self._pool_instances.put_nowait(instance_info)
                    except queue.Full:
                        instance_info.instance.release()
            except queue.Empty:
                break  # 超时

        # 如果等待后仍未获取到，创建一个新实例（超出池大小）
        logging.warning("Pool timeout, creating new beyond instance")
        try:
            instance_info = self._create_new_instance_info()
            instance_info.mark_in_use()  # 新实例直接标记为使用中
            with self._lock:
                self._total_active += 1
            logging.info(
                f"Created new beyond instance# {instance_info.instance_id} and marked as in-use (active: {self._total_active})"
            )
            return instance_info
        except Exception as e:
            logging.error(f"Failed to create new beyond instance: {e}")
            return None

    def put(self, instance_info: PoolInstanceInfo) -> None:
        """将实例归还到池中。"""
        if not instance_info:
            logging.warning("Attempted to put nil instance")
            return

        if self._stop_event.is_set():
            logging.warning("Pool is shutting down, releasing instance")
            instance_info.instance.release()
            return

        logging.info(f"Returning instance# {instance_info.instance_id} to pool({self._new_func})")

        if not instance_info.mark_available():
            logging.warning(
                f"Instance# {instance_info.instance_id} was not in use, cannot return to pool. This is unexpected."
            )
            return

        # 重置实例
        try:
            instance_info.instance.reset()
        except Exception as e:
            logging.warning(
                f"Failed to reset instance# {instance_info.instance_id} to pool({self._new_func}) err: {e}"
            )

        try:
            self._pool_instances.put_nowait(instance_info)
            with self._lock:
                self._total_active -= 1
            logging.info(
                f"Instance# {instance_info.instance_id} marked as available (active: {self._total_active}) to pool({self._new_func})"
            )
        except queue.Full:
            logging.warning(f"Pool queue full, releasing instance# {instance_info.instance_id}")
            instance_info.instance.release()

    def get_stats(self) -> Dict[str, Any]:
        """获取池的统计信息。"""
        with self._lock:
            return {
                "pool_size": self.pool_size,
                "total_instances": self._pool_instances.qsize(),
                "active_count": self._total_active,
                "total_created": self._total_created,
                "total_reused": self._total_reused,
            }

    def close(self) -> None:
        """关闭池，释放所有资源。"""
        logging.info("Pool Closing...")
        self._stop_event.set()

        # 清空队列并释放所有实例
        while True:
            try:
                instance_info = self._pool_instances.get_nowait()
                if instance_info:
                    instance_info.instance.release()
            except queue.Empty:
                break

        logging.info("Pool Closed")


"""
python -m src.common.pool_module
"""
if __name__ == "__main__":
    # --- 示例用法 ---
    class ExamplePoolableResource(IPoolInstance):
        """一个示例的可池化资源类。"""

        def __init__(self, resource_id: int):
            self.resource_id = resource_id
            self.data = f"Resource-{resource_id}-Data"
            logging.info(f"Created ExamplePoolableResource #{self.resource_id}")

        def reset(self) -> None:
            logging.info(f"Resetting ExamplePoolableResource #{self.resource_id}")
            # 模拟重置操作
            self.data = f"Resource-{self.resource_id}-Data-Reset"

        def release(self) -> None:
            logging.info(f"Releasing ExamplePoolableResource #{self.resource_id}")
            # 模拟释放资源操作
            self.data = None

    # 1. 创建函数
    def create_example_resource() -> IPoolInstance:
        """创建 ExamplePoolableResource 的工厂函数。"""
        # 这里可以使用一个静态计数器来模拟 ID 分配
        if not hasattr(create_example_resource, "counter"):
            create_example_resource.counter = 0
        create_example_resource.counter += 1
        return ExamplePoolableResource(create_example_resource.counter)

    # 2. 创建并初始化池
    pool = EngineProviderPool(pool_size=3, new_func=create_example_resource)
    if not pool.initialize():
        logging.error("Failed to initialize pool")
        exit(1)

    print("\n--- Pool Stats After Initialization ---")
    print(pool.get_stats())

    # 3. 获取实例
    print("\n--- Getting Instances ---")
    instance_info1 = pool.get()
    instance_info2 = pool.get()
    instance_info3 = pool.get()
    instance_info4 = pool.get()  # 这个会超出池大小创建

    print(
        f"Got instances: {instance_info1.instance_id}, {instance_info2.instance_id}, {instance_info3.instance_id}, {instance_info4.instance_id}"
    )

    print("\n--- Pool Stats After Getting ---")
    print(pool.get_stats())

    # 4. 模拟使用实例
    print(f"\n--- Using Instance 1 ---")
    resource1 = instance_info1.get_instance()
    print(f"Resource 1 Data: {resource1.data}")

    # 5. 归还实例
    print(f"\n--- Returning Instances 1, 2, 3 ---")
    pool.put(instance_info1)
    pool.put(instance_info2)
    pool.put(instance_info3)
    # 5.1. 归还超出池大小的实例
    print(f"\n--- Returning Beyond-Pool Instance 4 ---")
    pool.put(instance_info4)  # 这个会因为队列满而被释放

    print("\n--- Pool Stats After Returning 3 ---")
    print(pool.get_stats())

    # 6. 尝试获取一个刚归还的实例
    print(f"\n--- Getting Another Instance (should reuse) ---")
    instance_info5 = pool.get()
    print(f"Got instance: {instance_info5.instance_id} (Should be one of 1, 2, or 3)")

    print("\n--- Pool Stats After Re-getting ---")
    print(pool.get_stats())

    # 6.1. 已释放的不能放入
    print(f"\n--- Returning Beyond-Pool Instance 4 ---")
    pool.put(instance_info4)

    print("\n--- Final Pool Stats ---")
    print(pool.get_stats())

    # 7. 关闭池
    pool.close()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""isort:skip_file"""

import importlib
import os

from fairseq import registry
from fairseq.optim.lr_scheduler.fairseq_lr_scheduler import (  # noqa
    FairseqLRScheduler,
    LegacyFairseqLRScheduler,
)
from omegaconf import DictConfig


(
    build_lr_scheduler_,
    register_lr_scheduler,
    LR_SCHEDULER_REGISTRY,
    LR_SCHEDULER_DATACLASS_REGISTRY,
) = registry.setup_registry(
    "--lr-scheduler", base_class=FairseqLRScheduler, default="fixed"
)


def build_lr_scheduler(cfg: DictConfig, optimizer):
    return build_lr_scheduler_(cfg, optimizer)


# automatically import any Python files in the optim/lr_scheduler/ directory
for file in sorted(os.listdir(os.path.dirname(__file__))):
    if file.endswith(".py") and not file.startswith("_"):
        file_name = file[: file.find(".py")]
        importlib.import_module("fairseq.optim.lr_scheduler." + file_name)



import itertools
import os
import collections
import struct

import pyaudio
import pvporcupine
import wave


porcupine = None
paud = None
audio_stream = None

access_key = os.environ.get("PORCUPINE_ACCESS_KEY")


def get_defualt_instance():
    keywords = "hey google,terminator,americano,computer,pico clock,hey siri,grasshopper,alexa,jarvis,bumblebee,hey barista,blueberry,ok google,grapefruit,picovoice,porcupine".split(
        ","
    )
    print(keywords)
    porcupine = pvporcupine.create(access_key, keywords=keywords)
    return porcupine, keywords


def get_custom_instance():
    keywords = "小黑".split(",")
    porcupine = pvporcupine.create(
        access_key,
        keywords=keywords,
        model_path="./models/porcupine_params_zh.pv",
        keyword_paths=["./models/小黑_zh_mac_v3_0_0.ppn"],
    )
    return porcupine, keywords


def porcess_wakeword(porcupine, audio_buffer):
    # Removing the wake word from the recording
    samples_for_0_1_sec = int(porcupine.sample_rate * 0.1)
    start_index = max(0, len(audio_buffer) - samples_for_0_1_sec)
    print(f"start_index {start_index}, samples_for_0_1_sec {samples_for_0_1_sec}")
    temp_samples = collections.deque(itertools.islice(audio_buffer, start_index, None))
    audio_buffer.clear()
    audio_buffer.extend(temp_samples)
    return audio_buffer


FORMAT = pyaudio.paInt16
CHANNELS = 1
RATE = 16000


def export_wav(data, filename, sample_rate=RATE, channels=CHANNELS, sample_width=2):
    wf = wave.open(filename, "wb")
    wf.setnchannels(channels)
    wf.setsampwidth(sample_width)
    wf.setframerate(sample_rate)
    wf.writeframes(b"".join(data))
    wf.close()


try:
    # porcupine = get_defualt_instance()
    porcupine, keywords = get_custom_instance()
    print(porcupine.sample_rate, porcupine.frame_length)
    if porcupine.sample_rate != RATE:
        raise ValueError("Sample rate of your device is not %d Hz" % RATE)

    pre_recording_buffer_duration = 10.0
    maxlen = int((porcupine.sample_rate // porcupine.frame_length) * pre_recording_buffer_duration)
    print(f"audio_buffer maxlen: {maxlen}")
    # ring buffer
    audio_buffer = collections.deque(maxlen=maxlen)

    paud = pyaudio.PyAudio()
    audio_stream = paud.open(
        rate=porcupine.sample_rate,
        channels=1,
        format=pyaudio.paInt16,
        input=True,
        frames_per_buffer=porcupine.frame_length,
    )
    print("start recording")
    while True:
        read_audio_frames = audio_stream.read(porcupine.frame_length)
        # print(type(keyword), len(keyword))
        keyword = struct.unpack_from("h" * porcupine.frame_length, read_audio_frames)
        # print(type(keyword), len(keyword))
        keyword_index = porcupine.process(keyword)
        if keyword_index >= 0:
            print(
                f"index {keyword_index} {keywords[keyword_index]} hotword detected, audio_buffer length {len(audio_buffer)}"
            )

            audio_buffer = porcess_wakeword(porcupine, audio_buffer)
            print(f"audio_buffer length {len(audio_buffer)}")
            export_wav(
                audio_buffer,
                "./records/tmp_wakeword_porcupine.wav",
                sample_rate=porcupine.sample_rate,
                channels=CHANNELS,
                sample_width=paud.get_sample_size(FORMAT),
            )

        audio_buffer.append(read_audio_frames)
        # print(len(audio_buffer))

finally:
    if porcupine is not None:
        porcupine.delete()
    if audio_stream is not None:
        audio_stream.close()
    if paud is not None:
        paud.terminate()



import asyncio
import logging
from typing import Coroutine, Dict, Optional, Sequence
import uuid

from src.common.task_manager.base import Task, TaskManager


class AsyncioTaskManager(TaskManager):
    """
    use asyncio
    !TODO: @weedge
    - use [uvloop](https://github.com/MagicStack/uvloop) to speed up
    """

    def __init__(
        self, task_done_timeout: int = 5, loop: asyncio.AbstractEventLoop | None = None
    ) -> None:
        super().__init__(task_done_timeout=task_done_timeout)
        self._loop = loop or asyncio.get_event_loop()
        self._tasks: Dict[str, asyncio.Task] = {}

    async def run_task(self, target, name: str, tag: str, **kwargs):
        task_id = str(uuid.uuid4())

        async def wrapped_target():
            if asyncio.iscoroutinefunction(target):
                await target(**kwargs)
            else:
                # https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
                # Run in the default loop's executor(default thread pool)
                await self._loop.run_in_executor(None, target, **kwargs)

        task = self._loop.create_task(wrapped_target(), name=name)
        self._tasks[task_id] = Task(tid=task_id, name=name, tag=tag, task=task)

        return task_id

    async def cleanup(self):
        for task_id, val in self._tasks.items():
            task: asyncio.Task = val.task
            tag = val.tag
            try:
                if not task.done():
                    task.cancel()
                    try:
                        # wait task timeout
                        await asyncio.wait_for(task, self._task_done_timeout)
                    except asyncio.CancelledError:
                        logging.info(f"task_id:{task_id} tag:{tag} task: {task} cancelled")
                else:
                    logging.warning(f"task_id:{task_id} tag:{tag} task: {task} already completed")
            except Exception as e:
                logging.error(f"Error while cleaning up task {task_id}: {e}")

    def set_event_loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    def get_event_loop(self) -> asyncio.AbstractEventLoop:
        if not self._loop:
            raise Exception("TaskManager missing event loop, use TaskManager.set_event_loop().")
        return self._loop

    def create_task(self, coroutine: Coroutine, name: str) -> asyncio.Task:
        """
        Creates and schedules a new asyncio Task that runs the given coroutine.

        The task is added to a global set of created tasks.

        Args:
            loop (asyncio.AbstractEventLoop): The event loop to use for creating the task.
            coroutine (Coroutine): The coroutine to be executed within the task.
            name (str): The name to assign to the task for identification.

        Returns:
            asyncio.Task: The created task object.
        """

        async def run_coroutine():
            try:
                await coroutine
            except asyncio.CancelledError:
                logging.info(f"{name}: task cancelled")
                # Re-raise the exception to ensure the task is cancelled.
                raise
            except Exception as e:
                logging.exception(f"{name}: unexpected exception: {e}")

        if not self._loop:
            raise Exception("TaskManager missing event loop, use TaskManager.set_event_loop().")

        task = self._loop.create_task(run_coroutine())
        task.set_name(name)
        self._add_task(task)
        logging.info(f"{name}: task created")
        return task

    async def wait_for_task(self, task: asyncio.Task, timeout: Optional[float] = None):
        """Wait for an asyncio.Task to complete with optional timeout handling.

        This function awaits the specified asyncio.Task and handles scenarios for
        timeouts, cancellations, and other exceptions. It also ensures that the task
        is removed from the set of registered tasks upon completion or failure.

        Args:
            task (asyncio.Task): The asyncio Task to wait for.
            timeout (Optional[float], optional): The maximum number of seconds
                to wait for the task to complete. If None, waits indefinitely.
                Defaults to None.
        """
        name = task.get_name()
        try:
            if timeout:
                await asyncio.wait_for(task, timeout=timeout)
            else:
                await task
        except asyncio.TimeoutError:
            logging.warning(f"{name}: timed out waiting for task to finish")
        except asyncio.CancelledError:
            logging.info(f"{name}: unexpected task cancelled!")
            raise
        except Exception as e:
            logging.exception(f"{name}: unexpected exception while stopping task: {e}")
        finally:
            self._remove_task(task)

    async def cancel_task(self, task: asyncio.Task, timeout: Optional[float] = None):
        """Cancels the given asyncio Task and awaits its completion with an
        optional timeout.

        This function removes the task from the set of registered tasks upon
        completion or failure.

        Args:
            task (asyncio.Task): The task to be cancelled.
            timeout (Optional[float]): The optional timeout in seconds to wait for the task to cancel.

        """
        name = task.get_name()
        task.cancel()
        try:
            if timeout:
                await asyncio.wait_for(task, timeout=timeout)
            else:
                await task
        except asyncio.TimeoutError:
            logging.warning(f"{name}: timed out waiting for task to cancel")
        except asyncio.CancelledError:
            # Here are sure the task is cancelled properly.
            pass
        except Exception as e:
            logging.exception(f"{name}: unexpected exception while cancelling task: {e}")
        finally:
            self._remove_task(task)

    def current_tasks(self) -> Sequence[asyncio.Task]:
        """Returns the list of currently created/registered tasks."""
        return list(self._tasks.values())

    def _add_task(self, task: asyncio.Task):
        name = task.get_name()
        self._tasks[name] = task

    def _remove_task(self, task: asyncio.Task):
        name = task.get_name()
        try:
            del self._tasks[name]
        except KeyError as e:
            logging.warning(f"{name}: unable to remove task (already removed?): {e}")



import modal
import os

achatbot_version = os.getenv("ACHATBOT_VERSION", "0.0.26.post1")
# fastapi_webrtc_bots | fastapi_webrtc_single_bot server
SERVER_TAG = os.getenv("SERVER_TAG", "fastapi_webrtc_bots")
IMAGE_GPU = os.getenv("IMAGE_GPU", "A100-80GB")
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake")
    .pip_install("wheel", "openai", "qwen-omni-utils[decord]")
    .pip_install(
        [
            "achatbot["
            "fastapi_bot_server,"
            "livekit,livekit-api,daily,agora,"
            "silero_vad_analyzer,asr_processor,"
            "queue"
            f"]=={achatbot_version}",
        ],
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )
    .run_commands("pip install git+https://github.com/huggingface/transformers")
    .pip_install(
        "accelerate",
        "torch==2.7.0",
        "torchaudio==2.7.0",
        "torchvision==0.22.0",
        "soundfile==0.13.0",
        "librosa==0.11.0",
    )
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .env(
        {
            "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "ACHATBOT_PKG": "1",
            "SERVER_TAG": SERVER_TAG,
            "CONFIG_FILE": os.getenv(
                "CONFIG_FILE",
                "/root/.achatbot/config/bots/livekit_qwen3omni_vision_voice_bot.json",
            ),
            # https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#gpu-feature-list
        }
    )
)

# img = img.pip_install(
#    f"achatbot==0.0.26.dev20",
#    extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://test.pypi.org/simple/"),
# )

# ----------------------- app -------------------------------
app = modal.App("qwen3omni_bot")

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)
CONFIG_DIR = "/root/.achatbot/config"
config_vol = modal.Volume.from_name("config", create_if_missing=True)
RECORDS_DIR = "/root/.achatbot/records"
records_vol = modal.Volume.from_name("records", create_if_missing=True)
TORCH_CACHE_DIR = "/root/.cache/torch"
torch_cache_vol = modal.Volume.from_name("torch_cache", create_if_missing=True)


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.cls(
    image=img,
    gpu=os.getenv("IMAGE_GPU", None),
    secrets=[modal.Secret.from_name("achatbot")],
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
        CONFIG_DIR: config_vol,
        RECORDS_DIR: records_vol,
        TORCH_CACHE_DIR: torch_cache_vol,
    },
    cpu=2.0,
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
    # allow_concurrent_inputs=int(os.getenv("IMAGE_CONCURRENT_CN", "1")),
)
@modal.concurrent(max_inputs=int(os.getenv("IMAGE_CONCURRENT_CN", "1")))  # inputs per container
class Srv:
    @modal.enter()
    def enter(self):
        # run container runtime to enter when container is starting
        import subprocess
        import torch

        subprocess.run("nvidia-smi --version", shell=True)
        gpu_prop = None
        if torch.cuda.is_available():
            gpu_prop = torch.cuda.get_device_properties("cuda:0")
            print(gpu_prop)
            torch.multiprocessing.set_start_method("spawn", force=True)
        else:
            print("CUDA is not available.")

    @modal.asgi_app()
    def app(self):
        SERVER_TAG = os.getenv("SERVER_TAG", "fastapi_webrtc_bots")
        if SERVER_TAG == "fastapi_webrtc_single_bot":
            from achatbot.cmd.http.server.fastapi_room_bot_serve import app as fastapi_app

            print("run fastapi_room_bot_serve(single bot)")
        else:
            from achatbot.cmd.http.server.fastapi_daily_bot_serve import app as fastapi_app

            print("run fastapi_daily_bot_serve(multi bots)")

        return fastapi_app


"""
# 0. download models and assets
modal run src/download_models.py --repo-ids "Qwen/Qwen3-Omni-30B-A3B-Instruct"

# 1. run webrtc room http bots server

IMAGE_GPU=A100-80GB SERVER_TAG=fastapi_webrtc_bots \
    ACHATBOT_VERSION=0.0.26.post1 \
    modal serve src/fastapi_webrtc_qwen3omni_vision_voice_bot_serve.py

# 2. run webrtc room http signal bot server

modal volume create config
modal volume put config ./config/bots/livekit_qwen3omni_vision_voice_bot.json /bots/ -f


# run container with gpu
IMAGE_GPU=A100-80GB SERVER_TAG=fastapi_webrtc_bots \
    ACHATBOT_VERSION=0.0.26.post1 \
    CONFIG_FILE=/root/.achatbot/config/bots/livekit_qwen3omni_vision_voice_bot.json \
    modal serve src/fastapi_webrtc_qwen3omni_vision_voice_bot_serve.py

# cold start fastapi webrtc http server
curl -v -XGET "https://weedge--qwen3omni-bot-srv-app-dev.modal.run/health"

# run bot
curl -XPOST "https://weedge--qwen3omni-bot-srv-app-dev.modal.run/bot_join/chat-room/LivekitQwen3OmniVisionVoiceBot"


"""



from funasr import AutoModel
import soundfile
import typer

app = typer.Typer()


@app.command()
def online():
    """
    https://modelscope.cn/iic/damo/speech_fsmn_vad_zh-cn-16k-common-pytorch/summary
    """
    chunk_size = 200  # ms
    model = AutoModel(model="fsmn-vad", model_revision="v2.0.4")
    print(model.model)

    wav_file = f"{model.model_path}/example/vad_example.wav"
    speech, sample_rate = soundfile.read(wav_file)
    chunk_stride = int(chunk_size * sample_rate / 1000)

    cache = {}
    total_chunk_num = int(len((speech) - 1) / chunk_stride + 1)
    for i in range(total_chunk_num):
        speech_chunk = speech[i * chunk_stride : (i + 1) * chunk_stride]
        is_final = i == total_chunk_num - 1
        res = model.generate(
            input=speech_chunk,
            cache=cache,
            is_final=is_final,
            chunk_size=chunk_size,
            disable_pbar=True,
        )
        if len(res[0]["value"]):
            print(res)


@app.command()
def offline():
    """
    https://modelscope.cn/models/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch/summary
    """
    model = AutoModel(model="fsmn-vad", model_revision="v2.0.4")
    print(model.model_path)
    print(model.model)

    wav_file = f"{model.model_path}/example/vad_example.wav"
    res = model.generate(input=wav_file, disable_pbar=True)
    print(res)


"""
python demo/modelscope/fsmn_vad.py offline
python demo/modelscope/fsmn_vad.py online
"""
if __name__ == "__main__":
    app()

"""
FsmnVADStreaming(
  (encoder): FSMN(
    (in_linear1): AffineTransform(
      (linear): Linear(in_features=400, out_features=140, bias=True)
    )
    (in_linear2): AffineTransform(
      (linear): Linear(in_features=140, out_features=250, bias=True)
    )
    (relu): RectifiedLinear(
      (relu): ReLU()
      (dropout): Dropout(p=0.1, inplace=False)
    )
    (fsmn): FsmnStack(
      (0): BasicBlock(
        (linear): LinearTransform(
          (linear): Linear(in_features=250, out_features=128, bias=False)
        )
        (fsmn_block): FSMNBlock(
          (conv_left): Conv2d(128, 128, kernel_size=(20, 1), stride=(1, 1), groups=128, bias=False)
        )
        (affine): AffineTransform(
          (linear): Linear(in_features=128, out_features=250, bias=True)
        )
        (relu): RectifiedLinear(
          (relu): ReLU()
          (dropout): Dropout(p=0.1, inplace=False)
        )
      )
      (1): BasicBlock(
        (linear): LinearTransform(
          (linear): Linear(in_features=250, out_features=128, bias=False)
        )
        (fsmn_block): FSMNBlock(
          (conv_left): Conv2d(128, 128, kernel_size=(20, 1), stride=(1, 1), groups=128, bias=False)
        )
        (affine): AffineTransform(
          (linear): Linear(in_features=128, out_features=250, bias=True)
        )
        (relu): RectifiedLinear(
          (relu): ReLU()
          (dropout): Dropout(p=0.1, inplace=False)
        )
      )
      (2): BasicBlock(
        (linear): LinearTransform(
          (linear): Linear(in_features=250, out_features=128, bias=False)
        )
        (fsmn_block): FSMNBlock(
          (conv_left): Conv2d(128, 128, kernel_size=(20, 1), stride=(1, 1), groups=128, bias=False)
        )
        (affine): AffineTransform(
          (linear): Linear(in_features=128, out_features=250, bias=True)
        )
        (relu): RectifiedLinear(
          (relu): ReLU()
          (dropout): Dropout(p=0.1, inplace=False)
        )
      )
      (3): BasicBlock(
        (linear): LinearTransform(
          (linear): Linear(in_features=250, out_features=128, bias=False)
        )
        (fsmn_block): FSMNBlock(
          (conv_left): Conv2d(128, 128, kernel_size=(20, 1), stride=(1, 1), groups=128, bias=False)
        )
        (affine): AffineTransform(
          (linear): Linear(in_features=128, out_features=250, bias=True)
        )
        (relu): RectifiedLinear(
          (relu): ReLU()
          (dropout): Dropout(p=0.1, inplace=False)
        )
      )
    )
    (out_linear1): AffineTransform(
      (linear): Linear(in_features=250, out_features=140, bias=True)
    )
    (out_linear2): AffineTransform(
      (linear): Linear(in_features=140, out_features=248, bias=True)
    )
    (softmax): Softmax(dim=-1)
  )
)
"""


import os
import time
import logging
import asyncio
from typing import AsyncGenerator

import av
import cv2
import librosa
import numpy as np
import soundfile as sf
import torch
from apipeline.frames import Frame, StartFrame, EndFrame, CancelFrame

from achatbot.modules.avatar.musetalk import MusetalkAvatar
from achatbot.processors.avatar.base import AvatarProcessorBase, SegmentedAvatarProcessor
from achatbot.types.frames import AudioRawFrame, OutputAudioRawFrame, OutputImageRawFrame
from achatbot.types.avatar.musetalk import AvatarMuseTalkConfig
from achatbot.types.avatar import AvatarStatus, SpeechAudio
from achatbot.types.avatar.lite_avatar import AudioResult, VideoResult
from achatbot.processors.avatar.help.speech_audio_slicer import SpeechAudioSlicer
from achatbot.common.utils.audio_utils import bytes2NpArrayWith16


# class MusetalkAvatarProcessor(AvatarProcessorBase):
class MusetalkAvatarProcessor(SegmentedAvatarProcessor):
    def __init__(self, avatar: MusetalkAvatar, config: AvatarMuseTalkConfig, **kwargs):
        self._avatar = avatar
        self._config = config
        super().__init__(sample_rate=config.algo_audio_sample_rate, **kwargs)
        logging.info(f"init {__name__} init_config: {self._config}")

        # Internal algorithm sample rate, fixed at 16000
        self._algo_audio_sample_rate = config.algo_audio_sample_rate
        self._output_audio_sample_rate = config.output_audio_sample_rate

        # Internal queues
        self._audio_queue: asyncio.Queue = None  # Input audio queue
        self._whisper_queue: asyncio.Queue = None  # Whisper feature queue
        self._frame_id_queue: asyncio.Queue = None  # Frame ID allocation queue
        self._compose_queue: asyncio.Queue = None  # Frame composition queue
        self._output_queue: asyncio.Queue = None  # Output queue after composition

        # task
        self._feature_task: asyncio.Task = None
        self._frame_gen_task: asyncio.Task = None
        self._frame_collect_task: asyncio.Task = None
        self._compose_task: asyncio.Task = None
        self._session_running = False

        # warmup event
        self._feature_extractor_warmup_event = asyncio.Event()
        self._frame_generator_warmup_event = asyncio.Event()

        # Avatar status
        self._callback_avatar_status = AvatarStatus.LISTENING
        self._last_speech_id = None

        # Audio duration statistics
        self._first_add_audio_time = None
        self._audio_duration_sum = 0.0

        # Audio cache for each speech_id
        self._audio_cache = {}

        # audio input slice
        self._speech_audio_slicer: SpeechAudioSlicer = None

    async def start(self, frame: StartFrame):
        await super().start(frame)
        self._session_running = True
        self._init()
        await self._start_tasks()
        self._session_start_time = time.time()
        logging.info(f"{__name__} started")

    def _init(self):
        self._speech_audio_slicer = SpeechAudioSlicer(
            input_sample_rate=self._config.input_audio_sample_rate,  # input
            output_sample_rate=self._config.algo_audio_sample_rate,  # output for avatar input audio sample rate
            audio_slice_duration=self._config.input_audio_slice_duration,
        )

    async def _start_tasks(self):
        self._audio_queue = asyncio.Queue()  # Input audio queue
        self._whisper_queue = asyncio.Queue()  # Whisper feature queue
        self._frame_id_queue = asyncio.Queue()  # Frame ID allocation queue
        self._compose_queue = asyncio.Queue()  # Frame composition queue
        self._output_queue = asyncio.Queue()  # Output queue after composition

        self._feature_task: asyncio.Task = self.get_event_loop().create_task(
            self._feature_extractor_loop()
        )
        self._frame_gen_task: asyncio.Task = self.get_event_loop().create_task(
            self._frame_generator_loop()
        )
        self._frame_collect_task: asyncio.Task = self.get_event_loop().create_task(
            self._frame_collector_loop()
        )
        self._compose_task: asyncio.Task = self.get_event_loop().create_task(self._compose_loop())

        await self._feature_extractor_warmup_event.wait()
        self._feature_extractor_warmup_event.clear()

        await self._frame_generator_warmup_event.wait()
        self._frame_generator_warmup_event.clear()

    async def stop(self, frame: EndFrame):
        await super().stop(frame)
        await self._stop()

    async def cancle(self, frame: CancelFrame):
        await super().cancel(frame)
        await self._stop()

    async def _stop(self):
        logging.info(
            f"stop avatar processor {__name__}, totol session time {time.time() - self._session_start_time:.3f}",
        )
        self._feature_extractor_warmup_event.clear()
        self._frame_generator_warmup_event.clear()
        self._session_running = False
        if self._feature_task is not None:
            self._feature_task.cancel()
            await self._feature_task
            self._signal2img_task = None
        if self._frame_gen_task is not None:
            self._frame_gen_task.cancel()
            await self._frame_gen_task
            self._frame_gen_task = None
        if self._frame_collect_task is not None:
            self._frame_collect_task.cancel()
            await self._frame_collect_task
            self._frame_collect_task = None
        if self._compose_task is not None:
            self._compose_task.cancel()
            await self._compose_task
            self._compose_task = None
        logging.info(f"avatar processor {__name__} stopped")

    async def run_avatar(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        await self._add_audio(
            SpeechAudio(
                end_of_speech=True,
                speech_id="",
                sample_rate=frame.sample_rate,
                audio_data=frame.audio,
            )
        )

        # inner push/queue audio/image frame, yield None
        yield None

    async def _add_audio(self, speech_audio: SpeechAudio):
        """
        Add an audio segment to the processing queue. The segment length must not exceed 1 second. No resampling is performed here.
        Args:
            speech_audio (SpeechAudio): Audio segment to add.
        """
        if self._config.debug:
            now = time.time()
            # Record the first add_audio time
            if self._first_add_audio_time is None:
                self._first_add_audio_time = now
            # Calculate audio duration
            audio_len = len(speech_audio.audio_data)
            sample_rate = speech_audio.sample_rate
            audio_duration = audio_len / 4 / sample_rate  # float32, 4 bytes
            self._audio_duration_sum += audio_duration
            # Calculate cumulative interval
            total_interval = now - self._first_add_audio_time
            # Log output
            log_msg = (
                f"Received add_audio: speech_id={speech_audio.speech_id}, end_of_speech={speech_audio.end_of_speech}, "
                f"sample_rate={sample_rate}, audio_len={audio_len}, audio_duration={audio_duration:.3f}s, "
                f"cumulative_audio_duration={self._audio_duration_sum:.3f}s, total_interval={total_interval:.3f}s"
            )
            if self._audio_duration_sum < total_interval:
                logging.error(
                    log_msg + " [Cumulative audio duration < total interval, audio is too slow!]"
                )
            else:
                logging.info(log_msg)
            # Output cumulative duration and interval when end_of_speech is reached and reset
            if speech_audio.end_of_speech:
                logging.info(
                    f"[add_audio] speech_id={speech_audio.speech_id} segment cumulative_audio_duration: {self._audio_duration_sum:.3f}s, total_interval: {total_interval:.3f}s"
                )
                self._audio_duration_sum = 0.0
                self._first_add_audio_time = None

        audio_slices = self._speech_audio_slicer.get_speech_audio_slice(speech_audio)
        if self._config.debug:
            logging.info(f"audio_slices_len: {len(audio_slices)}")
        for audio_slice in audio_slices:
            if self._config.debug:
                logging.info(f"audio_slice: {str(audio_slice)}")

            audio_data = audio_slice.algo_audio_data

            # if len(audio_data) == 0:
            #    logging.error(f"Input audio is empty, speech_id={audio_slice.speech_id}")
            #    return

            ## Length check, process 1s  audio
            # if len(audio_data) > self._output_audio_sample_rate:
            #    logging.error(
            #        f"Audio segment too long: {len(audio_data)} > {self._algo_audio_sample_rate}, speech_id={audio_slice.speech_id}"
            #    )
            #    return

            assert speech_audio.sample_rate == self._output_audio_sample_rate, (
                f"{speech_audio.sample_rate=} != {self._output_audio_sample_rate=}"
            )

            await self._audio_queue.put(
                {
                    "audio_data": audio_data,  # Segment at algorithm sample rate (actually original segment)
                    "speech_id": audio_slice.speech_id,
                    "end_of_speech": audio_slice.end_of_speech,
                }
            )

    async def _feature_extractor_loop(self):
        """
        Worker thread for extracting audio features.
        """
        # warmup: ensure CUDA context and memory allocation (for whisper feature extraction)
        if torch.cuda.is_available():
            t0 = time.time()
            warmup_sr = 16000
            dummy_audio = np.zeros(warmup_sr, dtype=np.float32)
            await asyncio.to_thread(self._avatar.extract_whisper_feature, dummy_audio, warmup_sr)
            torch.cuda.synchronize()
            t1 = time.time()
            logging.info(
                f"_feature_extractor_loop whisper feature warmup once done, time: {(t1 - t0) * 1000:.1f} ms"
            )
            self._feature_extractor_warmup_event.set()

        while self._session_running is True:
            try:
                t_start = time.time()
                item = await asyncio.wait_for(self._audio_queue.get(), timeout=0.1)
                audio_data = item["audio_data"]
                speech_id = item["speech_id"]
                end_of_speech = item["end_of_speech"]
                fps = self._config.fps if hasattr(self._config, "fps") else 25

                segment = bytes2NpArrayWith16(audio_data)
                # Resample to algorithm sample rate
                if self._output_audio_sample_rate != self._algo_audio_sample_rate:
                    segment = librosa.resample(
                        segment,
                        orig_sr=self._output_audio_sample_rate,
                        target_sr=self._algo_audio_sample_rate,
                    )
                target_len = self._algo_audio_sample_rate  # 1 second
                if len(segment) < target_len:
                    segment = np.pad(segment, (0, target_len - len(segment)), mode="constant")

                # Feature extraction
                whisper_chunks = await asyncio.to_thread(
                    self._avatar.extract_whisper_feature, segment, self._algo_audio_sample_rate
                )

                audio_data = np.frombuffer(audio_data, dtype=np.float32)
                orig_audio_data_len = len(audio_data)
                orig_samples_per_frame = self._output_audio_sample_rate // fps
                actual_audio_len = orig_audio_data_len
                num_frames = int(np.ceil(actual_audio_len / orig_samples_per_frame))
                whisper_chunks = whisper_chunks[:num_frames]
                target_audio_len = num_frames * orig_samples_per_frame
                if len(audio_data) < target_audio_len:
                    audio_data = np.pad(
                        audio_data, (0, target_audio_len - len(audio_data)), mode="constant"
                    )
                else:
                    audio_data = audio_data[:target_audio_len]
                padded_audio_data_len = len(audio_data)
                await self._whisper_queue.put(
                    {
                        "whisper_chunks": whisper_chunks,
                        "speech_id": speech_id,
                        "end_of_speech": end_of_speech,
                        "audio_data": audio_data,
                    }
                )
                t_end = time.time()
                if self._config.debug:
                    logging.info(
                        f"[FEATURE_WORKER] speech_id={speech_id}, total_time={(t_end - t_start) * 1000:.1f}ms, whisper_chunks_frames={whisper_chunks.shape[0]}, audio_data_original_length={orig_audio_data_len}, audio_data_padded_length={padded_audio_data_len}, end_of_speech={end_of_speech}"
                    )
            except asyncio.CancelledError:
                logging.warning("feature_extractor_loop task cancelled")
                break
            except asyncio.TimeoutError:
                continue
            except Exception as e:
                logging.exception(f"feature_extractor_loop task error: {e}")
                continue
        logging.info("feature_extractor_loop stopped")

    def _get_audio_for_frame(self, audio_data, chunk_idx, num_chunks, orig_samples_per_frame):
        try:
            start_sample = chunk_idx * orig_samples_per_frame
            if chunk_idx == num_chunks - 1:
                audio_for_frame = audio_data[start_sample:]
                if len(audio_for_frame) < orig_samples_per_frame:
                    audio_for_frame = np.pad(
                        audio_for_frame,
                        (0, orig_samples_per_frame - len(audio_for_frame)),
                        mode="constant",
                    )
            else:
                end_sample = start_sample + orig_samples_per_frame
                audio_for_frame = audio_data[start_sample:end_sample]
            if len(audio_for_frame) < orig_samples_per_frame:
                logging.warning(
                    f"[AUDIO_PAD] Frame audio padding: {len(audio_for_frame)}->{orig_samples_per_frame}, chunk_idx={chunk_idx}"
                )
                audio_for_frame = np.pad(
                    audio_for_frame,
                    (0, orig_samples_per_frame - len(audio_for_frame)),
                    mode="constant",
                )
            return audio_for_frame
        except Exception as e:
            logging.error(
                f"[AUDIO_FOR_FRAME_ERROR] chunk_idx={chunk_idx}, error: {e}", exc_info=True
            )
            return np.zeros(orig_samples_per_frame, dtype=np.float32)

    async def _frame_generator_loop(self):
        """
        Generate speaking frames only, with rate control when queue buffer is full.
        Uses global frame_id allocation to ensure unique and continuous frame numbering for speaking frames.
        """
        from collections import namedtuple

        CurrentSpeechItem = namedtuple(
            "CurrentSpeechItem",
            ["whisper_chunks", "audio_data", "speech_id", "end_of_speech", "num_chunks"],
        )
        current_item = None
        chunk_idx = 0
        fps = self._config.fps
        orig_samples_per_frame = int(self._output_audio_sample_rate / fps)
        batch_size = self._config.batch_size  # Can be adjusted based on actual needs
        # max_speaking_buffer = batch_size * 5  # Maximum length of speaking frame buffer

        # warmup, ensure CUDA context and memory allocation
        if torch.cuda.is_available():
            t0 = time.time()
            # Regular batch_size warmup
            dummy_whisper = torch.zeros(
                batch_size, 50, 384, device=self._avatar.device, dtype=self._avatar.weight_dtype
            )
            await asyncio.to_thread(self._avatar.generate_frames, dummy_whisper, 0, batch_size)
            # Remainder batch_size warmup (only when there's a remainder)
            remain = fps % batch_size
            if remain > 0:
                dummy_whisper_remain = torch.zeros(
                    remain, 50, 384, device=self._avatar.device, dtype=self._avatar.weight_dtype
                )
                await asyncio.to_thread(
                    self._avatar.generate_frames, dummy_whisper_remain, 0, remain
                )
            torch.cuda.synchronize()
            t1 = time.time()
            logging.info(f"_frame_generator_loop self-warmup done, time: {(t1 - t0) * 1000:.1f} ms")
            self._frame_generator_warmup_event.set()

        while self._session_running is True:
            try:
                if current_item is None:
                    item = await asyncio.wait_for(self._whisper_queue.get(), timeout=0.1)
                    fetched_chunks = item["whisper_chunks"]
                    num_fetched_chunks = (
                        fetched_chunks.shape[0]
                        if isinstance(fetched_chunks, torch.Tensor) and fetched_chunks.ndim > 0
                        else 0
                    )
                    if num_fetched_chunks > 0:
                        current_item = CurrentSpeechItem(
                            whisper_chunks=fetched_chunks,
                            audio_data=item["audio_data"],
                            speech_id=item["speech_id"],
                            end_of_speech=item["end_of_speech"],
                            num_chunks=num_fetched_chunks,
                        )
                        chunk_idx = 0

                # Batch inference for speaking frames
                if current_item is not None and chunk_idx < current_item.num_chunks:
                    remain = current_item.num_chunks - chunk_idx
                    cur_batch = min(batch_size, remain)
                    batch_start_time = time.time()
                    # Get frame_id from frame_id queue
                    t1 = time.time()
                    frame_ids = []
                    for _ in range(cur_batch):
                        try:
                            item = self._frame_id_queue.get_nowait()
                            frame_ids.append(item)
                        except asyncio.QueueEmpty:
                            await asyncio.sleep(0.1)
                            continue
                    cost = time.time() - t1
                    print(f"{cost=}")
                    whisper_batch = current_item.whisper_chunks[chunk_idx : chunk_idx + cur_batch]
                    try:
                        recon_idx_list = await asyncio.to_thread(
                            self._avatar.generate_frames, whisper_batch, frame_ids[0], cur_batch
                        )
                    except Exception as e:
                        logging.error(
                            f"[GEN_FRAME_ERROR] frame_id={frame_ids[0]}, speech_id={current_item.speech_id}, error: {e}",
                            exc_info=True,
                        )
                        recon_idx_list = [
                            (np.zeros((256, 256, 3), dtype=np.uint8), frame_ids[0] + i)
                            for i in range(cur_batch)
                        ]
                    batch_end_time = time.time()
                    if self._config.debug:
                        logging.info(
                            f"[FRAME_GEN] Generated speaking frame: speech_id={current_item.speech_id}, chunk_idx={chunk_idx}, cur_batch={cur_batch}, batch_time={(batch_end_time - batch_start_time) * 1000:.1f}ms"
                        )
                    for i in range(cur_batch):
                        recon, idx = recon_idx_list[i]
                        audio = self._get_audio_for_frame(
                            current_item.audio_data,
                            chunk_idx + i,
                            current_item.num_chunks,
                            orig_samples_per_frame,
                        )
                        is_last = chunk_idx + i == current_item.num_chunks - 1
                        eos = current_item.end_of_speech and is_last
                        # Directly put into compose queue, no longer controlled by fps
                        compose_item = {
                            "recon": recon,
                            "idx": idx,
                            "speech_id": current_item.speech_id,
                            "avatar_status": AvatarStatus.SPEAKING,
                            "end_of_speech": eos,
                            "audio_segment": audio,
                            "frame_id": idx,
                            "timestamp": time.time(),
                        }
                        await self._compose_queue.put(compose_item)
                    if chunk_idx + cur_batch >= current_item.num_chunks:
                        current_item = None
                        chunk_idx = 0
                    else:
                        chunk_idx += cur_batch
                    continue
            except asyncio.CancelledError:
                logging.warning("frame_generator_loop task cancelled")
                break
            except asyncio.TimeoutError:
                continue
            except Exception as e:
                logging.exception(f"frame_generator_loop task error: {e}")
                continue
        logging.info("frame_generator loop stopped")

    async def _compose_loop(self):
        """
        Responsible for executing res2combined and putting the synthesis results into _output_queue
        """
        while self._session_running is True:
            try:
                item = await asyncio.wait_for(self._compose_queue.get(), timeout=0.1)
                recon = item["recon"]
                idx = item["idx"]
                frame = await asyncio.to_thread(self._avatar.res2combined, recon, idx)
                item["frame"] = frame
                await self._output_queue.put(item)
            except asyncio.CancelledError:
                logging.warning("compose_loop task cancelled")
                break
            except asyncio.TimeoutError:
                continue
            except Exception as e:
                logging.exception(f"compose_loop task error: {e}")
                continue
        logging.info("compose_loop stopped")

    async def _frame_collector_loop(self):
        """
        Collector strictly outputs at fps, with frame numbers matching the frame_id assigned by the inference thread.
        """
        fps = self._config.fps
        frame_interval = 1.0 / fps
        start_time = time.perf_counter()
        local_frame_id = 0
        last_active_speech_id = None
        last_speaking = False
        last_end_of_speech = False
        current_speech_id = None
        while self._session_running is True:
            try:
                # Control fps
                target_time = start_time + local_frame_id * frame_interval
                now = time.perf_counter()
                sleep_time = target_time - now
                if sleep_time > 0.002:
                    await asyncio.sleep(sleep_time - 0.001)
                # while time.perf_counter() < target_time:
                #    pass

                # Record the start time for profiling
                t_frame_start = time.perf_counter()
                # Allocate frame_id
                await self._frame_id_queue.put(local_frame_id)

                if self._output_queue.empty():
                    frame = self._avatar.generate_idle_frame(local_frame_id)
                    speech_id = last_active_speech_id
                    avatar_status = AvatarStatus.LISTENING
                    end_of_speech = False
                    frame_timestamp = time.time()
                    audio_segment = None
                else:
                    output_item = self._output_queue.get_nowait()
                    frame = output_item["frame"]
                    speech_id = output_item["speech_id"]
                    avatar_status = output_item["avatar_status"]
                    end_of_speech = output_item["end_of_speech"]
                    frame_timestamp = output_item.get("timestamp", None)
                    audio_segment = output_item["audio_segment"]

                # Notify video
                video_frame = av.VideoFrame.from_ndarray(frame, format="bgr24")
                video_result = VideoResult(
                    video_frame=video_frame,
                    speech_id=speech_id,
                    avatar_status=avatar_status,
                    end_of_speech=end_of_speech,
                )
                await self._callback_image(video_result)

                # Logging logic
                is_idle = avatar_status == AvatarStatus.LISTENING and speech_id is None
                is_speaking = avatar_status == AvatarStatus.SPEAKING
                is_end_of_speech = bool(end_of_speech)
                if self._config.debug:
                    if is_speaking:
                        # First speaking frame
                        if speech_id != current_speech_id:
                            logging.info(
                                f"[SPEAKING_FRAME][START] frame_id={local_frame_id}, speech_id={speech_id}, status={avatar_status}, end_of_speech={end_of_speech}, video_timestamp={frame_timestamp}"
                            )
                            current_speech_id = speech_id
                        # Last speaking frame
                        if is_end_of_speech:
                            logging.info(
                                f"[SPEAKING_FRAME][END] frame_id={local_frame_id}, speech_id={speech_id}, status={avatar_status}, end_of_speech={end_of_speech}, video_timestamp={frame_timestamp}"
                            )
                            current_speech_id = None
                        # Middle speaking frame
                        if not is_end_of_speech and (speech_id == current_speech_id):
                            logging.info(
                                f"[SPEAKING_FRAME] frame_id={local_frame_id}, speech_id={speech_id}, status={avatar_status}, end_of_speech={end_of_speech}, video_timestamp={frame_timestamp}"
                            )
                    elif is_idle and last_speaking:
                        if last_end_of_speech:
                            logging.info(
                                f"[IDLE_FRAME] Start after speaking: frame_id={local_frame_id}, status={avatar_status}"
                            )
                        else:
                            logging.warning(
                                f"[IDLE_FRAME] Inserted idle during speaking: frame_id={local_frame_id}"
                            )
                else:
                    if is_speaking and speech_id != current_speech_id:
                        logging.info(
                            f"[SPEAKING_FRAME] Start: frame_id={local_frame_id}, speech_id={speech_id}"
                        )
                        current_speech_id = speech_id
                    # Last speaking frame
                    if is_speaking and is_end_of_speech:
                        logging.info(
                            f"[SPEAKING_FRAME] End: frame_id={local_frame_id}, speech_id={speech_id}, end_of_speech=True"
                        )
                        current_speech_id = None
                    # Idle frame: distinguish between inserted idle during speaking and first idle after speaking
                    if is_idle and last_speaking:
                        if last_end_of_speech:
                            logging.info(
                                f"[IDLE_FRAME] Start after speaking: frame_id={local_frame_id}"
                            )
                        else:
                            logging.warning(
                                f"[IDLE_FRAME] Inserted idle during speaking: frame_id={local_frame_id}"
                            )

                # Audio related
                audio_len = len(audio_segment) if audio_segment is not None else 0
                if audio_len > 0:
                    audio_np = np.asarray(audio_segment, dtype=np.float32)
                    if audio_np.ndim == 1:
                        audio_np = audio_np[np.newaxis, :]
                    audio_frame = av.AudioFrame.from_ndarray(audio_np, format="flt", layout="mono")
                    audio_frame.sample_rate = self._output_audio_sample_rate
                    audio_result = AudioResult(
                        audio_frame=audio_frame, speech_id=speech_id, end_of_speech=end_of_speech
                    )
                    if speech_id not in self._audio_cache:
                        self._audio_cache[speech_id] = []
                    self._audio_cache[speech_id].append(
                        audio_np[0] if audio_np.ndim == 2 else audio_np
                    )
                    audio_len_sum = (
                        sum([len(seg) for seg in self._audio_cache[speech_id]])
                        / self._output_audio_sample_rate
                    )
                    if self._config.debug:
                        logging.info(
                            f"[AUDIO_FRAME] frame_id={local_frame_id}, speech_id={speech_id}, end_of_speech={end_of_speech}, audio_timestamp={frame_timestamp}, Cumulative audio duration={audio_len_sum:.3f}s"
                        )
                    await self._callback_audio(audio_result)

                # Status switching etc.
                if end_of_speech:
                    logging.info(f"Status change: SPEAKING -> LISTENING, speech_id={speech_id}")
                    try:
                        if getattr(self._config, "debug_save_handler_audio", False):
                            all_audio = np.concatenate(self._audio_cache[speech_id], axis=-1)
                            save_dir = "logs/audio_segments"
                            os.makedirs(save_dir, exist_ok=True)
                            wav_path = os.path.join(save_dir, f"{speech_id}_all.wav")
                            sf.write(
                                wav_path,
                                all_audio,
                                self._output_audio_sample_rate,
                                subtype="PCM_16",
                            )
                            logging.info(f"[AUDIO_FRAME] saved full wav: {wav_path}")
                    except Exception as e:
                        logging.error(f"[AUDIO_FRAME] save full wav error: {e}")
                    del self._audio_cache[speech_id]
                    self._notify_status_change(speech_id, AvatarStatus.LISTENING)

                t_frame_end = time.perf_counter()
                if self._config.debug and (t_frame_end - t_frame_start > frame_interval):
                    logging.warning(
                        f"[PROFILE] frame_id={local_frame_id} total={t_frame_end - t_frame_start:.4f}s (>{frame_interval:.4f}s)"
                    )

                local_frame_id += 1
                last_speaking = is_speaking
                last_end_of_speech = is_end_of_speech
            except asyncio.CancelledError:
                logging.warning("frame_collector_loop task cancelled")
                break
            except Exception as e:
                logging.exception(f"frame_collector_loop task error: {e}")
                continue
        logging.info("frame_collector_loop loop ended")

    async def _callback_image(self, image_result: VideoResult):
        video_frame: av.VideoFrame = image_result.video_frame
        img = video_frame.to_image()
        if self._session_running:
            await self.queue_frame(
                OutputImageRawFrame(
                    image=img.tobytes(),
                    size=img.size,
                    format=img.format,
                    mode=img.mode,
                )
            )

    async def _callback_audio(self, audio_result: AudioResult):
        audio_frame: av.AudioFrame = audio_result.audio_frame
        if self._session_running:
            await self.queue_frame(
                OutputAudioRawFrame(
                    audio=audio_frame.to_ndarray().tobytes(),
                    sample_rate=audio_frame.sample_rate,
                )
            )

    def _notify_status_change(self, speech_id: str, status: AvatarStatus):
        pass



import sys
import h5py
import soundfile as sf
import fire
import math
import yamlargparse
import numpy as np
from torch.utils.data import DataLoader
import torch
from utils.utils import parse_config_or_kwargs
from utils.dataset import DiarizationDataset
from models.models import TransformerDiarization
from scipy.signal import medfilt
from sklearn.cluster import AgglomerativeClustering
from scipy.spatial import distance
from utils.kaldi_data import KaldiData


def get_cl_sil(args, acti, cls_num):
    n_chunks = len(acti)
    mean_acti = np.array([np.mean(acti[i], axis=0)
                         for i in range(n_chunks)]).flatten()
    n = args.num_speakers
    sil_spk_th = args.sil_spk_th

    cl_lst = []
    sil_lst = []
    for chunk_idx in range(n_chunks):
        if cls_num is not None:
            if args.num_speakers > cls_num:
                mean_acti_bi = np.array([mean_acti[n * chunk_idx + s_loc_idx]
                                        for s_loc_idx in range(n)])
                min_idx = np.argmin(mean_acti_bi)
                mean_acti[n * chunk_idx + min_idx] = 0.0

        for s_loc_idx in range(n):
            a = n * chunk_idx + (s_loc_idx + 0) % n
            b = n * chunk_idx + (s_loc_idx + 1) % n
            if mean_acti[a] > sil_spk_th and mean_acti[b] > sil_spk_th:
                cl_lst.append((a, b))
            else:
                if mean_acti[a] <= sil_spk_th:
                    sil_lst.append(a)

    return cl_lst, sil_lst


def clustering(args, svec, cls_num, ahc_dis_th, cl_lst, sil_lst):
    org_svec_len = len(svec)
    svec = np.delete(svec, sil_lst, 0)

    # update cl_lst idx
    _tbl = [i - sum(sil < i for sil in sil_lst) for i in range(org_svec_len)]
    cl_lst = [(_tbl[_cl[0]], _tbl[_cl[1]]) for _cl in cl_lst]

    distMat = distance.cdist(svec, svec, metric='euclidean')
    for cl in cl_lst:
        distMat[cl[0], cl[1]] = args.clink_dis
        distMat[cl[1], cl[0]] = args.clink_dis

    clusterer = AgglomerativeClustering(
            n_clusters=cls_num,
            affinity='precomputed',
            linkage='average',
            distance_threshold=ahc_dis_th)
    clusterer.fit(distMat)

    if cls_num is not None:
        print("oracle n_clusters is known")
    else:
        print("oracle n_clusters is unknown")
        print("estimated n_clusters by constraind AHC: {}"
              .format(len(np.unique(clusterer.labels_))))
        cls_num = len(np.unique(clusterer.labels_))

    sil_lab = cls_num
    insert_sil_lab = [sil_lab for i in range(len(sil_lst))]
    insert_sil_lab_idx = [sil_lst[i] - i for i in range(len(sil_lst))]
    print("insert_sil_lab : {}".format(insert_sil_lab))
    print("insert_sil_lab_idx : {}".format(insert_sil_lab_idx))
    clslab = np.insert(clusterer.labels_,
                       insert_sil_lab_idx,
                       insert_sil_lab).reshape(-1, args.num_speakers)
    print("clslab : {}".format(clslab))

    return clslab, cls_num


def merge_act_max(act, i, j):
    for k in range(len(act)):
        act[k, i] = max(act[k, i], act[k, j])
        act[k, j] = 0.0
    return act


def merge_acti_clslab(args, acti, clslab, cls_num):
    sil_lab = cls_num
    for i in range(len(clslab)):
        _lab = clslab[i].reshape(-1, 1)
        distM = distance.cdist(_lab, _lab, metric='euclidean').astype(np.int64)
        for j in range(len(distM)):
            distM[j][:j] = -1
        idx_lst = np.where(np.count_nonzero(distM == 0, axis=1) > 1)
        merge_done = []
        for j in idx_lst[0]:
            for k in (np.where(distM[j] == 0))[0]:
                if j != k and clslab[i, j] != sil_lab and k not in merge_done:
                    print("merge : (i, j, k) == ({}, {}, {})".format(i, j, k))
                    acti[i] = merge_act_max(acti[i], j, k)
                    clslab[i, k] = sil_lab
                    merge_done.append(j)

    return acti, clslab


def stitching(args, acti, clslab, cls_num):
    n_chunks = len(acti)
    s_loc = args.num_speakers
    sil_lab = cls_num
    s_tot = max(cls_num, s_loc-1)

    # Extend the max value of s_loc_idx to s_tot+1
    add_acti = []
    for chunk_idx in range(n_chunks):
        zeros = np.zeros((len(acti[chunk_idx]), s_tot+1))
        if s_tot+1 > s_loc:
            zeros[:, :-(s_tot+1-s_loc)] = acti[chunk_idx]
        else:
            zeros = acti[chunk_idx]
        add_acti.append(zeros)
    acti = np.array(add_acti)

    out_chunks = []
    for chunk_idx in range(n_chunks):
        # Make sloci2lab_dct.
        # key: s_loc_idx
        # value: estimated label by clustering or sil_lab
        cls_set = set()
        for s_loc_idx in range(s_tot+1):
            cls_set.add(s_loc_idx)

        sloci2lab_dct = {}
        for s_loc_idx in range(s_tot+1):
            if s_loc_idx < s_loc:
                sloci2lab_dct[s_loc_idx] = clslab[chunk_idx][s_loc_idx]
                if clslab[chunk_idx][s_loc_idx] in cls_set:
                    cls_set.remove(clslab[chunk_idx][s_loc_idx])
                else:
                    if clslab[chunk_idx][s_loc_idx] != sil_lab:
                        raise ValueError
            else:
                sloci2lab_dct[s_loc_idx] = list(cls_set)[s_loc_idx-s_loc]

        # Sort by label value
        sloci2lab_lst = sorted(sloci2lab_dct.items(), key=lambda x: x[1])

        # Select sil_lab_idx
        sil_lab_idx = None
        for idx_lab in sloci2lab_lst:
            if idx_lab[1] == sil_lab:
                sil_lab_idx = idx_lab[0]
                break
        if sil_lab_idx is None:
            raise ValueError

        # Get swap_idx
        # [idx of label(0), idx of label(1), ..., idx of label(s_tot)]
        swap_idx = [sil_lab_idx for j in range(s_tot+1)]
        for lab in range(s_tot+1):
            for idx_lab in sloci2lab_lst:
                if lab == idx_lab[1]:
                    swap_idx[lab] = idx_lab[0]

        print("swap_idx {}".format(swap_idx))
        swap_acti = acti[chunk_idx][:, swap_idx]
        swap_acti = np.delete(swap_acti, sil_lab, 1)
        out_chunks.append(swap_acti)

    return out_chunks


def prediction(num_speakers, net, wav_list, chunk_len_list):
    acti_lst = []
    svec_lst = []
    len_list = []

    with torch.no_grad():
        for wav, chunk_len in zip(wav_list, chunk_len_list):
            wav = wav.to('cuda')
            outputs = net.batch_estimate(torch.unsqueeze(wav, 0))
            ys = outputs[0]

            for i in range(num_speakers):
                spkivecs = outputs[i+1]
                svec_lst.append(spkivecs[0].cpu().detach().numpy())

            acti = ys[0][-chunk_len:].cpu().detach().numpy()
            acti_lst.append(acti)
            len_list.append(chunk_len)

    acti_arr = np.concatenate(acti_lst, axis=0)  # totol_len x num_speakers
    svec_arr = np.stack(svec_lst)                # (chunk_num x num_speakers) x emb_dim
    len_arr = np.array(len_list)                 # chunk_num

    return acti_arr, svec_arr, len_arr

def cluster(args, conf, acti_arr, svec_arr, len_arr):

    acti_list = []
    n_chunks = len_arr.shape[0]
    start = 0
    for i in range(n_chunks):
        chunk_len = len_arr[i]
        acti_list.append(acti_arr[start: start+chunk_len])
        start += chunk_len
    acti = np.array(acti_list)
    svec = svec_arr

    # initialize clustering setting
    cls_num = None
    ahc_dis_th = args.ahc_dis_th
    # Get cannot-link index list and silence index list
    cl_lst, sil_lst = get_cl_sil(args, acti, cls_num)

    n_samples = n_chunks * args.num_speakers - len(sil_lst)
    min_n_samples = 2
    if cls_num is not None:
        min_n_samples = cls_num

    if n_samples >= min_n_samples:
        # clustering (if cls_num is None, update cls_num)
        clslab, cls_num =\
             clustering(args, svec, cls_num, ahc_dis_th, cl_lst, sil_lst)
        # merge
        acti, clslab = merge_acti_clslab(args, acti, clslab, cls_num)
        # stitching
        out_chunks = stitching(args, acti, clslab, cls_num)
    else:
        out_chunks = acti

    outdata = np.vstack(out_chunks)
    # Saving the resuts
    return outdata

def make_rttm(args, conf, cluster_data):
    args.frame_shift = conf['model']['frame_shift']
    args.subsampling = conf['model']['subsampling']
    args.sampling_rate = conf['dataset']['sampling_rate']

    with open(args.out_rttm_file, 'w') as wf:
        a = np.where(cluster_data > args.threshold, 1, 0)
        if args.median > 1:
            a = medfilt(a, (args.median, 1))
        for spkid, frames in enumerate(a.T):
            frames = np.pad(frames, (1, 1), 'constant')
            changes, = np.where(np.diff(frames, axis=0) != 0)
            fmt = "SPEAKER {:s} 1 {:7.2f} {:7.2f} <NA> <NA> {:s} <NA>"
            for s, e in zip(changes[::2], changes[1::2]):
                print(fmt.format(
                    args.session,
                    s * args.frame_shift * args.subsampling / args.sampling_rate,
                    (e - s) * args.frame_shift * args.subsampling / args.sampling_rate,
                    args.session + "_" + str(spkid)), file=wf)

def main(args):
    conf = parse_config_or_kwargs(args.config_path)
    num_speakers = conf['dataset']['num_speakers']
    args.num_speakers = num_speakers

    # Prepare model
    model_parameter_dict = torch.load(args.model_init)['model']
    model_all_n_speakers = model_parameter_dict["embed.weight"].shape[0]
    conf['model']['all_n_speakers'] = model_all_n_speakers
    net = TransformerDiarization(**conf['model'])
    net.load_state_dict(model_parameter_dict, strict=False)
    net.eval()
    net = net.to("cuda")

    audio, sr = sf.read(args.wav_path, dtype="float32")
    audio_len = audio.shape[0]
    chunk_size, frame_shift, subsampling = conf['dataset']['chunk_size'], conf['model']['frame_shift'], conf['model']['subsampling']
    scale_ratio = int(frame_shift * subsampling)
    chunk_audio_size = chunk_size * scale_ratio
    wav_list, chunk_len_list = [], []
    for i in range(0, math.ceil(1.0 * audio_len / chunk_audio_size)):
        start, end = i*chunk_audio_size, (i+1)*chunk_audio_size
        if end > audio_len:
            chunk_len_list.append(int((audio_len-start) / scale_ratio))
            end = audio_len
            start = max(0, audio_len - chunk_audio_size)
        else:
            chunk_len_list.append(chunk_size)
        wav_list.append(audio[start:end])
    wav_list = [torch.from_numpy(wav).float() for wav in wav_list]

    acti_arr, svec_arr, len_arr = prediction(num_speakers, net, wav_list, chunk_len_list)
    cluster_data = cluster(args, conf, acti_arr, svec_arr, len_arr)
    make_rttm(args, conf, cluster_data)

if __name__ == '__main__':
    parser = yamlargparse.ArgumentParser(description='decoding')
    parser.add_argument('--wav_path',
                        help='the input wav path',
                        default="tmp/mix_0000496.wav")
    parser.add_argument('--config_path',
                        help='config file path',
                        default="config/infer_est_nspk1.yaml")
    parser.add_argument('--model_init',
                        help='model initialize path',
                        default="")
    parser.add_argument('--sil_spk_th', default=0.05, type=float)
    parser.add_argument('--ahc_dis_th', default=1.0, type=float)
    parser.add_argument('--clink_dis', default=1.0e+4, type=float)
    parser.add_argument('--session', default='Anonymous', help='the name of the output speaker')
    parser.add_argument('--out_rttm_file', default='out.rttm', help='the output rttm file')
    parser.add_argument('--threshold', default=0.4, type=float)
    parser.add_argument('--median', default=25, type=int)


    args = parser.parse_args()
    main(args)
